import hashlib
import logging
import os
import shutil

try:
    import rucio.common
    from rucio.client import Client
    from rucio.client.downloadclient import DownloadClient
    from rucio.client.uploadclient import UploadClient

    from .rucio_extra_clients import (
        DeleteClient,
        InPlaceIngestClient,
    )
except ImportError:
    Client = None

try:
    from galaxy.authnz.util import provider_name_to_backend
except ImportError:
    provider_name_to_backend = None  # type: ignore[assignment, unused-ignore]

from galaxy.exceptions import (
    ObjectInvalid,
    ObjectNotFound,
)
from galaxy.util import (
    directory_hash_id,
    string_as_bool,
    umask_fix_perms,
    unlink,
)
from ._caching_base import CachingConcreteObjectStore
from .caching import (
    enable_cache_monitor,
    parse_caching_config_dict_from_xml,
)

log = logging.getLogger(__name__)

NO_RUCIO_ERROR_MESSAGE = (
    "ObjectStore configured to use Rucio, but no rucio-clients dependency available."
    "Please install and properly configure rucio-clients or modify Object "
    "Store configuration."
)


def _config_xml_error(tag):
    msg = f"No {tag} element in config XML tree"
    raise Exception(msg)


def _config_dict_error(key):
    msg = f"No {key} key in config dictionary"
    raise Exception(msg)


def _encode_key(input_string):
    input_bytes = input_string.encode("utf-8")
    sha256_hash = hashlib.sha256(input_bytes)
    hashed_string = sha256_hash.hexdigest()
    return hashed_string


def parse_config_xml(config_xml):
    try:
        cache_dict = parse_caching_config_dict_from_xml(config_xml)

        attrs = ("type", "path")
        e_xml = config_xml.findall("extra_dir")
        if not e_xml:
            _config_xml_error("extra_dir")
        extra_dirs = [{k: e.get(k) for k in attrs} for e in e_xml]

        attrs_schemes = ("rse", "scheme", "ignore_checksum")
        e_xml = config_xml.findall("rucio_download_scheme")
        rucio_download_schemes = []
        if e_xml:
            rucio_download_schemes = [{k: e.get(k) for k in attrs_schemes} for e in e_xml]

        oidc_provider = config_xml.findtext("oidc_provider", None)
        enable_cache_mon = string_as_bool(config_xml.findtext("enable_cache_monitor", "False"))

        e_xml = config_xml.findall("rucio_upload_scheme")
        if e_xml:
            rucio_upload_rse_name = e_xml[0].get("rse", None)
            rucio_upload_scheme = e_xml[0].get("scheme", None)
            rucio_scope = e_xml[0].get("scope", None)
            rucio_register_only = string_as_bool(e_xml[0].get("register_only", "False"))
        else:
            rucio_upload_rse_name = None
            rucio_upload_scheme = None
            rucio_scope = None
            rucio_register_only = False
            oidc_provider = None

        e_xml = config_xml.findall("rucio_auth")
        if not e_xml:
            _config_xml_error("rucio_auth")
        rucio_account = e_xml[0].get("account", None)
        rucio_auth_host = e_xml[0].get("host", None)
        rucio_username = e_xml[0].get("username", None)
        rucio_password = e_xml[0].get("password", None)
        rucio_auth_type = e_xml[0].get("type", "userpass")

        e_xml = config_xml.findall("rucio_connection")
        if not e_xml:
            _config_xml_error("rucio_connection")
        rucio_host = e_xml[0].get("host", None)

        rucio_dict = {
            "upload_rse_name": rucio_upload_rse_name,
            "upload_scheme": rucio_upload_scheme,
            "scope": rucio_scope,
            "register_only": rucio_register_only,
            "download_schemes": rucio_download_schemes,
            "account": rucio_account,
            "auth_host": rucio_auth_host,
            "username": rucio_username,
            "password": rucio_password,
            "auth_type": rucio_auth_type,
            "host": rucio_host,
        }

        return {
            "cache": cache_dict,
            "rucio": rucio_dict,
            "extra_dirs": extra_dirs,
            "oidc_provider": oidc_provider,
            "enable_cache_monitor": enable_cache_mon,
        }
    except Exception:
        # Toss it back up after logging, we can't continue loading at this point.
        log.exception("Malformed Rucio ObjectStore Configuration XML -- unable to continue.")
        raise


class RucioBroker:
    def __init__(self, rucio_config):
        self._temp_file_name = None
        self.config = rucio_config
        self.upload_scheme = rucio_config["upload_scheme"]
        self.upload_rse_name = rucio_config["upload_rse_name"]
        self.scope = rucio_config["scope"]
        self.register_only = rucio_config["register_only"]
        self.download_schemes = rucio_config["download_schemes"]
        if Client is None:
            raise Exception(NO_RUCIO_ERROR_MESSAGE)
        rucio.common.utils.PREFERRED_CHECKSUM = "md5"

    def get_rucio_client(self):
        client = Client(
            rucio_host=self.config["host"],
            auth_host=self.config["auth_host"],
            account=self.config["account"],
            auth_type=self.config["auth_type"],
            creds={"username": self.config["username"], "password": self.config["username"]},
        )
        return client

    def get_rucio_upload_client(self, auth_token=None):
        client = self.get_rucio_client()
        uc = UploadClient(_client=client)
        uc.auth_token = auth_token
        return uc

    def get_rucio_download_client(self, auth_token=None):
        client = self.get_rucio_client()
        dc = DownloadClient(client=client)
        dc.auth_token = auth_token
        return dc

    def get_rucio_ingest_client(self, auth_token=None):
        client = self.get_rucio_client()
        ic = InPlaceIngestClient(_client=client)
        ic.auth_token = auth_token
        return ic

    def get_rucio_delete_client(self, auth_token=None):
        client = self.get_rucio_client()
        ic = DeleteClient(_client=client)
        ic.auth_token = auth_token
        return ic

    def register(self, key, source_path):
        key = _encode_key(key)
        item = {
            "path": source_path,
            "rse": self.upload_rse_name,
            "did_scope": self.scope,
            "did_name": key,
            "pfn": f"file://localhost/{source_path}",
        }
        items = [item]
        self.get_rucio_ingest_client().ingest(items)

    def upload(self, key, source_path):
        key = _encode_key(key)
        item = {
            "path": source_path,
            "rse": self.upload_rse_name,
            "did_scope": self.scope,
            "did_name": key,
            "force_scheme": self.upload_scheme,
        }
        items = [item]
        self.get_rucio_upload_client().upload(items)

    def download(self, key, dest_path, auth_token):
        key = _encode_key(key)
        base_dir = os.path.dirname(dest_path)
        dids = [{"scope": self.scope, "name": key}]
        try:
            repl = next(self.get_rucio_client().list_replicas(dids))["rses"].keys()
            item = None
            for rse_scheme in self.download_schemes:
                if rse_scheme["rse"] in repl:
                    item = {
                        "did": f"{self.scope}:{key}",
                        "force_scheme": rse_scheme["scheme"],
                        "rse": rse_scheme["rse"],
                        "base_dir": base_dir,
                        "check_local_with_filesize_only": string_as_bool(rse_scheme["ignore_checksum"]),
                        "ignore_checksum": string_as_bool(rse_scheme["ignore_checksum"]),
                        "no_subdir": True,
                    }
                    break
            if item is None:
                item = {
                    "did": f"{self.scope}:{key}",
                    "base_dir": base_dir,
                    "no_subdir": True,
                }
            items = [item]
            download_client = self.get_rucio_download_client(auth_token=auth_token)
            download_client.download_dids(items)
        except Exception as e:
            log.exception(f"Cannot download file: {str(e)}")
            return False
        return True

    def data_object_exists(self, key):
        key = _encode_key(key)
        dids = [{"scope": self.scope, "name": key}]
        try:
            repl = next(self.get_rucio_client().list_replicas(dids))
            return "AVAILABLE" in repl["states"].values()
        except Exception:
            return False

    def get_size(self, key):
        key = _encode_key(key)
        dids = [{"scope": self.scope, "name": key}]
        try:
            repl = next(self.get_rucio_client().list_replicas(dids))
            return repl["bytes"]
        except Exception:
            return 0

    def delete(self, key, auth_token):
        key = _encode_key(key)
        try:
            items = [{"did": {"scope": self.scope, "name": key}}]
            self.get_rucio_delete_client(auth_token=auth_token).delete(items, self.download_schemes, True)
        except Exception as e:
            log.exception(f"Cannot delete file: {e}")
            return False
        return True


class RucioObjectStore(CachingConcreteObjectStore):
    """
    Object store implementation that uses ORNL remote data broker.

    This implementation should be considered beta and may be dropped from
    Galaxy at some future point or significantly modified.
    """

    store_type = "rucio"

    def to_dict(self):
        rval = super().to_dict()
        rval["rucio"] = self.rucio_config
        rval["cache"] = self.cache_config
        rval["oidc_provider"] = self.oidc_provider
        rval["enable_cache_monitor"] = self.enable_cache_monitor
        return rval

    def __init__(self, config, config_dict):
        super().__init__(config, config_dict)
        self.rucio_config = config_dict.get("rucio") or {}

        self.oidc_provider = config_dict.get("oidc_provider", None)
        self.rucio_broker = RucioBroker(self.rucio_config)
        cache_dict = config_dict.get("cache") or {}
        self.enable_cache_monitor, self.cache_monitor_interval = enable_cache_monitor(config, config_dict)

        self.cache_size = cache_dict.get("size") or self.config.object_store_cache_size
        self.staging_path = cache_dict.get("path") or self.config.object_store_cache_path
        self.cache_updated_data = cache_dict.get("cache_updated_data", True)
        self.cache_config = cache_dict
        self._initialize()

    def _initialize(self):
        self._ensure_staging_path_writable()
        self._start_cache_monitor_if_needed()

    def _pull_into_cache(self, rel_path, auth_token):
        log.debug("rucio _pull_into_cache: %s", rel_path)
        # Ensure the cache directory structure exists (e.g., dataset_#_files/)
        rel_path_dir = os.path.dirname(rel_path)
        if not os.path.exists(self._get_cache_path(rel_path_dir)):
            os.makedirs(self._get_cache_path(rel_path_dir), exist_ok=True)
        # Now pull in the file
        dest = self._get_cache_path(rel_path)
        file_ok = self.rucio_broker.download(rel_path, dest, auth_token)
        self._fix_permissions(self._get_cache_path(rel_path_dir))
        return file_ok

    def _fix_file_permissions(self, path):
        umask_fix_perms(path, self.config.umask, 0o666)

    def _fix_permissions(self, rel_path):
        """Set permissions on rel_path"""
        for basedir, _, files in os.walk(rel_path):
            umask_fix_perms(basedir, self.config.umask, 0o777)
            for filename in files:
                path = os.path.join(basedir, filename)
                # Ignore symlinks
                if os.path.islink(path):
                    continue
                umask_fix_perms(path, self.config.umask, 0o666)

    # "interfaces to implement"

    def _exists(self, obj, **kwargs):
        rel_path = self._construct_path(obj, **kwargs)
        log.debug("rucio _exists: %s", rel_path)

        dir_only = kwargs.get("dir_only", False)
        base_dir = kwargs.get("base_dir", None)

        # Check cache and rucio
        if self._in_cache(rel_path) or (not dir_only and self.rucio_broker.data_object_exists(rel_path)):
            return True

        # dir_only does not get synced so shortcut the decision
        if dir_only and base_dir:
            # for JOB_WORK directory
            if not os.path.exists(rel_path):
                os.makedirs(rel_path, exist_ok=True)
            return True
        return False

    @classmethod
    def parse_xml(cls, config_xml):
        return parse_config_xml(config_xml)

    def _create(self, obj, **kwargs):
        if not self._exists(obj, **kwargs):
            # Pull out locally used fields
            extra_dir = kwargs.get("extra_dir", None)
            extra_dir_at_root = kwargs.get("extra_dir_at_root", False)
            dir_only = kwargs.get("dir_only", False)
            alt_name = kwargs.get("alt_name", None)

            # Construct hashed path
            rel_path = os.path.join(*directory_hash_id(self._get_object_id(obj)))

            # Optionally append extra_dir
            if extra_dir is not None:
                if extra_dir_at_root:
                    rel_path = os.path.join(extra_dir, rel_path)
                else:
                    rel_path = os.path.join(rel_path, extra_dir)

            # Create given directory in cache
            cache_dir = os.path.join(self.staging_path, rel_path)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir, exist_ok=True)

            if not dir_only:
                rel_path = os.path.join(rel_path, alt_name if alt_name else f"dataset_{self._get_object_id(obj)}.dat")
                # need this line to set the dataset filename, not sure how this is done - filesystem is monitored?
                open(os.path.join(self.staging_path, rel_path), "w").close()
            log.debug("rucio _create: %s", rel_path)
        return self

    def _size(self, obj, **kwargs):
        rel_path = self._construct_path(obj, **kwargs)
        log.debug("rucio _size: %s", rel_path)

        if self._in_cache(rel_path):
            try:
                size = os.path.getsize(self._get_cache_path(rel_path))
            except OSError as ex:
                log.info("Could not get size of file '%s' in local cache, will try iRODS. Error: %s", rel_path, ex)
            if size != 0:
                return size
        if self._exists(obj, **kwargs):
            return self._get_remote_size(rel_path)
        log.warning("Did not find dataset '%s', returning 0 for size", rel_path)
        return 0

    def _get_remote_size(self, rel_path):
        return self.rucio_broker.get_size(rel_path)

    def _delete(self, obj, entire_dir=False, **kwargs):
        rel_path = self._construct_path(obj, **kwargs)
        extra_dir = kwargs.get("extra_dir", None)
        base_dir = kwargs.get("base_dir", None)
        dir_only = kwargs.get("dir_only", False)
        obj_dir = kwargs.get("obj_dir", False)
        log.debug("rucio _delete: %s", rel_path)
        auth_token = self._get_token(**kwargs)

        try:
            # Remove temporary data in JOB_WORK directory
            if base_dir and dir_only and obj_dir:
                shutil.rmtree(os.path.abspath(rel_path))
                return True

            # Delete from cache first
            if entire_dir and extra_dir:
                shutil.rmtree(self._get_cache_path(rel_path), ignore_errors=True)
            else:
                unlink(self._get_cache_path(rel_path), ignore_errors=True)

            # Delete from rucio as well
            if self.rucio_broker.data_object_exists(rel_path):
                self.rucio_broker.delete(rel_path, auth_token)
                return True
        except OSError:
            log.exception("%s delete error", self._get_filename(obj, **kwargs))
        return False

    def _get_token(self, **kwargs):
        auth_token = kwargs.get("auth_token", None)
        if auth_token:
            return auth_token

        arg_user = kwargs.get("user", None)
        try:
            if not arg_user:
                trans = kwargs.get("trans", None)
                user = trans.user
            else:
                user = arg_user
            backend = provider_name_to_backend(self.oidc_provider)
            tokens = user.get_oidc_tokens(backend)
            return tokens["id"]
        except Exception as e:
            log.debug("Failed to get auth token: %s", e)
            return None

    def _get_filename(self, obj, **kwargs):
        base_dir = kwargs.get("base_dir", None)
        dir_only = kwargs.get("dir_only", False)
        auth_token = self._get_token(**kwargs)
        rel_path = self._construct_path(obj, **kwargs)
        sync_cache = kwargs.get("sync_cache", True)

        log.debug("rucio _get_filename: %s", rel_path)

        # for JOB_WORK directory
        if base_dir and dir_only:
            return os.path.abspath(rel_path)

        cache_path = self._get_cache_path(rel_path)
        if not sync_cache:
            return cache_path

        in_cache = self._in_cache(rel_path)
        size_in_cache = 0
        if in_cache:
            size_in_cache = os.path.getsize(self._get_cache_path(rel_path))

        # return path if we do not need to update cache
        if in_cache and dir_only:
            return cache_path
        # something is already in cache
        elif in_cache:
            size_in_rdb = self.rucio_broker.get_size(rel_path)
            # same size as in  rucio, or empty file in rucio - do not pull
            if size_in_cache == size_in_rdb or size_in_rdb == 0:
                return cache_path

        # Check if the file exists in persistent storage and, if it does, pull it into cache
        if self._exists(obj, **kwargs):
            if dir_only:  # Directories do not get pulled into cache
                return cache_path
            else:
                if self._pull_into_cache(rel_path, auth_token):
                    return cache_path
        raise ObjectNotFound(f"objectstore.get_filename, no cache_path: {obj}, kwargs: {kwargs}")

    def _register_file(self, rel_path, file_name):
        if file_name is None:
            file_name = self._get_cache_path(rel_path)
            if not os.path.islink(file_name):
                raise ObjectInvalid(
                    "rucio objectstore._register_file, rucio_register_only is set, but file in cache is not a link "
                )
        if os.path.islink(file_name):
            file_name = os.readlink(file_name)
        self.rucio_broker.register(rel_path, file_name)
        log.debug("rucio _register_file: %s", file_name)
        return

    def _update_from_file(self, obj, file_name=None, create=False, **kwargs):
        rel_path = self._construct_path(obj, **kwargs)
        log.debug("rucio _update_from_file: %s", rel_path)

        if not create:
            log.warning(
                "rucio objectstore.update_from_file, file update without create, will fail if file already in Rucio"
            )

        if self.rucio_config["register_only"]:
            self._register_file(rel_path, file_name)
            return

        # Choose whether to use the dataset file itself or an alternate file
        if file_name:
            source_file = os.path.abspath(file_name)
            # Copy into cache
            cache_file = self._get_cache_path(rel_path)
            try:
                if source_file != cache_file and self.cache_updated_data:
                    try:
                        shutil.copy2(source_file, cache_file)
                    except OSError:
                        os.makedirs(os.path.dirname(cache_file))
                        shutil.copy2(source_file, cache_file)
                self._fix_file_permissions(cache_file)
                source_file = cache_file
            except OSError:
                log.exception("Trouble copying source file '%s' to cache '%s'", source_file, cache_file)
        else:
            source_file = self._get_cache_path(rel_path)

        # Update the file on rucio
        self.rucio_broker.upload(rel_path, source_file)

    def _get_store_usage_percent(self, **kwargs):
        log.debug("rucio _get_store_usage_percent, not implemented yet")
        return 0.0

    def _get_object_url(self, obj, extra_dir=None, extra_dir_at_root=False, alt_name=None):
        log.debug("rucio _get_object_url")
        return None

    def __build_kwargs(self, obj, **kwargs):
        kwargs["object_id"] = obj.id
        return kwargs

    def shutdown(self):
        self._shutdown_cache_monitor()



from typing import (
    Any,
    Dict,
    List,
    Optional,
    Type,
    Union,
)

from pydantic import RootModel
from typing_extensions import Literal

from galaxy.objectstore.badges import (
    BadgeDict,
    StoredBadgeDict,
)
from galaxy.util.config_templates import (
    EnvironmentDict,
    expand_raw_config,
    MarkdownContent,
    populate_default_variables,
    SecretsDict,
    StrictModel,
    TemplateEnvironmentEntry,
    TemplateExpansion,
    TemplateSecret,
    TemplateVariable,
    TemplateVariableType,
    TemplateVariableValueType,
    UserDetailsDict,
)

ObjectStoreTemplateVariableType = TemplateVariableType
ObjectStoreTemplateVariableValueType = TemplateVariableValueType
ObjectStoreTemplateType = Literal["aws_s3", "azure_blob", "boto3", "disk", "generic_s3"]


class S3AuthTemplate(StrictModel):
    access_key: Union[str, TemplateExpansion]
    secret_key: Union[str, TemplateExpansion]


class S3Auth(StrictModel):
    access_key: str
    secret_key: str


class S3BucketTemplate(StrictModel):
    name: Union[str, TemplateExpansion]
    use_reduced_redundancy: Optional[Union[bool, TemplateExpansion]] = None


class S3Bucket(StrictModel):
    name: str
    use_reduced_redundancy: Optional[bool] = None


BadgeList = Optional[List[StoredBadgeDict]]


class AwsS3ObjectStoreTemplateConfiguration(StrictModel):
    type: Literal["aws_s3"]
    auth: S3AuthTemplate
    bucket: S3BucketTemplate
    badges: BadgeList = None
    template_start: Optional[str] = None
    template_end: Optional[str] = None


class AwsS3ObjectStoreConfiguration(StrictModel):
    type: Literal["aws_s3"]
    auth: S3Auth
    bucket: S3Bucket
    badges: BadgeList = None


class AzureAuthTemplate(StrictModel):
    account_name: Union[str, TemplateExpansion]
    account_key: Union[str, TemplateExpansion]


class AzureAuth(StrictModel):
    account_name: str
    account_key: str


class AzureContainerTemplate(StrictModel):
    name: Union[str, TemplateExpansion]


class AzureContainer(StrictModel):
    name: str


class AzureTransferTemplate(StrictModel):
    max_concurrency: Optional[Union[int, TemplateExpansion]] = None
    download_max_concurrency: Optional[Union[int, TemplateExpansion]] = None
    upload_max_concurrency: Optional[Union[int, TemplateExpansion]] = None
    max_single_put_size: Optional[Union[int, TemplateExpansion]] = None
    max_single_get_size: Optional[Union[int, TemplateExpansion]] = None
    max_block_size: Optional[Union[int, TemplateExpansion]] = None


class AzureObjectStoreTemplateConfiguration(StrictModel):
    type: Literal["azure_blob"]
    auth: AzureAuthTemplate
    container: AzureContainerTemplate
    transfer: Optional[AzureTransferTemplate] = None
    badges: BadgeList = None
    template_start: Optional[str] = None
    template_end: Optional[str] = None


class AzureTransfer(StrictModel):
    max_concurrency: Optional[int] = None
    download_max_concurrency: Optional[int] = None
    upload_max_concurrency: Optional[int] = None
    max_single_put_size: Optional[int] = None
    max_single_get_size: Optional[int] = None
    max_block_size: Optional[int] = None


class AzureObjectStoreConfiguration(StrictModel):
    type: Literal["azure_blob"]
    auth: AzureAuth
    container: AzureContainer
    transfer: Optional[AzureTransfer] = None
    badges: BadgeList = None


class Boto3BucketTemplate(StrictModel):
    name: Union[str, TemplateExpansion]


class Boto3ConnectionTemplate(StrictModel):
    endpoint_url: Union[str, TemplateExpansion]
    region: Optional[Union[str, TemplateExpansion]] = None


class Boto3TransferTemplate(StrictModel):
    use_threads: Optional[Union[bool, TemplateExpansion]] = None
    multipart_threshold: Optional[Union[int, TemplateExpansion]] = None
    max_concurrency: Optional[Union[int, TemplateExpansion]] = None
    multipart_chunksize: Optional[Union[int, TemplateExpansion]] = None
    num_download_attempts: Optional[Union[int, TemplateExpansion]] = None
    max_io_queue: Optional[Union[int, TemplateExpansion]] = None
    io_chunksize: Optional[Union[int, TemplateExpansion]] = None
    max_bandwidth: Optional[Union[int, TemplateExpansion]] = None
    download_use_threads: Optional[Union[bool, TemplateExpansion]] = None
    download_multipart_threshold: Optional[Union[int, TemplateExpansion]] = None
    download_max_concurrency: Optional[Union[int, TemplateExpansion]] = None
    download_multipart_chunksize: Optional[Union[int, TemplateExpansion]] = None
    download_num_download_attempts: Optional[Union[int, TemplateExpansion]] = None
    download_max_io_queue: Optional[Union[int, TemplateExpansion]] = None
    download_io_chunksize: Optional[Union[int, TemplateExpansion]] = None
    download_max_bandwidth: Optional[Union[int, TemplateExpansion]] = None
    upload_use_threads: Optional[Union[bool, TemplateExpansion]] = None
    upload_multipart_threshold: Optional[Union[int, TemplateExpansion]] = None
    upload_max_concurrency: Optional[Union[int, TemplateExpansion]] = None
    upload_multipart_chunksize: Optional[Union[int, TemplateExpansion]] = None
    upload_num_download_attempts: Optional[Union[int, TemplateExpansion]] = None
    upload_max_io_queue: Optional[Union[int, TemplateExpansion]] = None
    upload_io_chunksize: Optional[Union[int, TemplateExpansion]] = None
    upload_max_bandwidth: Optional[Union[int, TemplateExpansion]] = None


class Boto3ObjectStoreTemplateConfiguration(StrictModel):
    type: Literal["boto3"]
    auth: S3AuthTemplate
    bucket: Boto3BucketTemplate
    connection: Optional[Boto3ConnectionTemplate] = None
    transfer: Optional[Boto3TransferTemplate] = None
    badges: BadgeList = None
    template_start: Optional[str] = None
    template_end: Optional[str] = None


class Boto3Bucket(StrictModel):
    name: str


class Boto3Connection(StrictModel):
    endpoint_url: str
    region: Optional[str] = None


class Boto3Transfer(StrictModel):
    use_threads: Optional[bool] = None
    multipart_threshold: Optional[int] = None
    max_concurrency: Optional[int] = None
    multipart_chunksize: Optional[int] = None
    num_download_attempts: Optional[int] = None
    max_io_queue: Optional[int] = None
    io_chunksize: Optional[int] = None
    max_bandwidth: Optional[int] = None
    download_use_threads: Optional[bool] = None
    download_multipart_threshold: Optional[int] = None
    download_max_concurrency: Optional[int] = None
    download_multipart_chunksize: Optional[int] = None
    download_num_download_attempts: Optional[int] = None
    download_max_io_queue: Optional[int] = None
    download_io_chunksize: Optional[int] = None
    download_max_bandwidth: Optional[int] = None
    upload_use_threads: Optional[bool] = None
    upload_multipart_threshold: Optional[int] = None
    upload_max_concurrency: Optional[int] = None
    upload_multipart_chunksize: Optional[int] = None
    upload_num_download_attempts: Optional[int] = None
    upload_max_io_queue: Optional[int] = None
    upload_io_chunksize: Optional[int] = None
    upload_max_bandwidth: Optional[int] = None


class Boto3ObjectStoreConfiguration(StrictModel):
    type: Literal["boto3"]
    auth: S3Auth
    bucket: Boto3Bucket
    connection: Optional[Boto3Connection] = None
    transfer: Optional[Boto3Transfer] = None
    badges: BadgeList = None


class DiskObjectStoreTemplateConfiguration(StrictModel):
    type: Literal["disk"]
    files_dir: Union[str, TemplateExpansion]
    badges: BadgeList = None
    template_start: Optional[str] = None
    template_end: Optional[str] = None


class DiskObjectStoreConfiguration(StrictModel):
    type: Literal["disk"]
    files_dir: str
    badges: BadgeList = None


class S3ConnectionTemplate(StrictModel):
    host: Union[str, TemplateExpansion]
    port: Union[int, TemplateExpansion]
    is_secure: Optional[Union[bool, TemplateExpansion]] = True
    conn_path: Optional[Union[str, TemplateExpansion]] = ""


class S3Connection(StrictModel):
    host: str
    port: int
    is_secure: bool = True
    conn_path: str = ""


class GenericS3ObjectStoreTemplateConfiguration(StrictModel):
    type: Literal["generic_s3"]
    auth: S3AuthTemplate
    bucket: S3BucketTemplate
    connection: S3ConnectionTemplate
    badges: BadgeList = None
    template_start: Optional[str] = None
    template_end: Optional[str] = None


class GenericS3ObjectStoreConfiguration(StrictModel):
    type: Literal["generic_s3"]
    auth: S3Auth
    bucket: S3Bucket
    connection: S3Connection
    badges: BadgeList = None


ObjectStoreTemplateConfiguration = Union[
    AwsS3ObjectStoreTemplateConfiguration,
    Boto3ObjectStoreTemplateConfiguration,
    GenericS3ObjectStoreTemplateConfiguration,
    DiskObjectStoreTemplateConfiguration,
    AzureObjectStoreTemplateConfiguration,
]
ObjectStoreConfiguration = Union[
    AwsS3ObjectStoreConfiguration,
    Boto3ObjectStoreConfiguration,
    DiskObjectStoreConfiguration,
    AzureObjectStoreConfiguration,
    GenericS3ObjectStoreConfiguration,
]


ObjectStoreTemplateVariable = TemplateVariable
ObjectStoreTemplateSecret = TemplateSecret


class ObjectStoreTemplateBase(StrictModel):
    """Version of ObjectStoreTemplate we can send to the UI/API.

    The configuration key in the child type may have secretes
    and shouldn't be exposed over the API - at least to non-admins.
    """

    id: str
    name: Optional[str]
    description: Optional[MarkdownContent]
    # The UI should just show the most recent version but allow
    # admins to define newer versions with new parameterizations
    # and keep old versions in template catalog for backward compatibility
    # for users with existing stores of that template.
    version: int = 0
    # Like with multiple versions, allow admins to deprecate a
    # template by hiding but keep it in the catalog for backward
    # compatibility for users with existing stores of that template.
    hidden: bool = False
    variables: Optional[List[TemplateVariable]] = None
    secrets: Optional[List[TemplateSecret]] = None


class ObjectStoreTemplateSummary(ObjectStoreTemplateBase):
    badges: List[BadgeDict]
    type: ObjectStoreTemplateType


class ObjectStoreTemplate(ObjectStoreTemplateBase):
    configuration: ObjectStoreTemplateConfiguration
    environment: Optional[List[TemplateEnvironmentEntry]] = None


ObjectStoreTemplateCatalog = RootModel[List[ObjectStoreTemplate]]


class ObjectStoreTemplateSummaries(RootModel):
    root: List[ObjectStoreTemplateSummary]


def template_to_configuration(
    template: ObjectStoreTemplate,
    variables: Dict[str, ObjectStoreTemplateVariableValueType],
    secrets: SecretsDict,
    user_details: UserDetailsDict,
    environment: EnvironmentDict,
) -> ObjectStoreConfiguration:
    configuration_template = template.configuration
    populate_default_variables(template.variables, variables)
    raw_config = expand_raw_config(configuration_template, variables, secrets, user_details, environment)
    return to_configuration_object(raw_config)


TypesToConfigurationClasses: Dict[ObjectStoreTemplateType, Type[ObjectStoreConfiguration]] = {
    "aws_s3": AwsS3ObjectStoreConfiguration,
    "boto3": Boto3ObjectStoreConfiguration,
    "generic_s3": GenericS3ObjectStoreConfiguration,
    "azure_blob": AzureObjectStoreConfiguration,
    "disk": DiskObjectStoreConfiguration,
}


def to_configuration_object(configuration_dict: Dict[str, Any]) -> ObjectStoreConfiguration:
    if "type" not in configuration_dict:
        raise KeyError("Configuration objects require an object store 'type' key, none found.")
    object_store_type = configuration_dict["type"]
    if object_store_type not in TypesToConfigurationClasses:
        raise ValueError(f"Unknown object store type found in raw configuration dictionary ({object_store_type}).")
    return TypesToConfigurationClasses[object_store_type](**configuration_dict)



"""
Object Store plugin for Cloud storage.
"""

import logging
import os
import os.path

from ._caching_base import CachingConcreteObjectStore
from ._util import UsesAxel
from .caching import enable_cache_monitor
from .s3 import parse_config_xml

try:
    from cloudbridge.factory import (
        CloudProviderFactory,
        ProviderList,
    )
    from cloudbridge.interfaces.exceptions import InvalidNameException
except ImportError:
    CloudProviderFactory = None
    ProviderList = None

log = logging.getLogger(__name__)

NO_CLOUDBRIDGE_ERROR_MESSAGE = (
    "Cloud ObjectStore is configured, but no CloudBridge dependency available."
    "Please install CloudBridge or modify ObjectStore configuration."
)


class Cloud(CachingConcreteObjectStore, UsesAxel):
    """
    Object store that stores objects as items in an cloud storage. A local
    cache exists that is used as an intermediate location for files between
    Galaxy and the cloud storage.
    """

    store_type = "cloud"

    def __init__(self, config, config_dict):
        super().__init__(config, config_dict)

        bucket_dict = config_dict["bucket"]
        cache_dict = config_dict.get("cache") or {}
        self.enable_cache_monitor, self.cache_monitor_interval = enable_cache_monitor(config, config_dict)

        self.provider = config_dict["provider"]
        self.credentials = config_dict["auth"]
        self.bucket_name = bucket_dict.get("name")
        self.use_rr = bucket_dict.get("use_reduced_redundancy", False)
        self.max_chunk_size = bucket_dict.get("max_chunk_size", 250)

        self.cache_size = cache_dict.get("size") or self.config.object_store_cache_size
        self.staging_path = cache_dict.get("path") or self.config.object_store_cache_path
        self.cache_updated_data = cache_dict.get("cache_updated_data", True)

        self._initialize()

    def _initialize(self):
        if CloudProviderFactory is None:
            raise Exception(NO_CLOUDBRIDGE_ERROR_MESSAGE)

        self.conn = self._get_connection(self.provider, self.credentials)
        self.bucket = self._get_bucket(self.bucket_name)
        self._ensure_staging_path_writable()
        self._start_cache_monitor_if_needed()
        self._init_axel()

    @staticmethod
    def _get_connection(provider, credentials):
        log.debug(f"Configuring `{provider}` Connection")
        if provider == "aws":
            config = {"aws_access_key": credentials["access_key"], "aws_secret_key": credentials["secret_key"]}
            if "region" in credentials:
                config["aws_region_name"] = credentials["region"]
            connection = CloudProviderFactory().create_provider(ProviderList.AWS, config)
        elif provider == "azure":
            config = {
                "azure_subscription_id": credentials["subscription_id"],
                "azure_client_id": credentials["client_id"],
                "azure_secret": credentials["secret"],
                "azure_tenant": credentials["tenant"],
            }
            connection = CloudProviderFactory().create_provider(ProviderList.AZURE, config)
        elif provider == "google":
            config = {"gcp_service_creds_file": credentials["credentials_file"]}
            connection = CloudProviderFactory().create_provider(ProviderList.GCP, config)
        else:
            raise Exception(f"Unsupported provider `{provider}`.")

        # Ideally it would be better to assert if the connection is
        # authorized to perform operations required by ObjectStore
        # before returning it (and initializing ObjectStore); hence
        # any related issues can be handled properly here, and ObjectStore
        # can "trust" the connection is established.
        #
        # However, the mechanism implemented in Cloudbridge to assert if
        # a user/service is authorized to perform an operation, assumes
        # the user/service is granted with an elevated privileges, such
        # as admin/owner-level access to all resources. For a detailed
        # discussion see:
        #
        # https://github.com/CloudVE/cloudbridge/issues/135
        #
        # Hence, if a resource owner wants to only authorize Galaxy to r/w
        # a bucket/container on the provider, but does not allow it to access
        # other resources, Cloudbridge may fail asserting credentials.
        # For instance, to r/w an Amazon S3 bucket, the resource owner
        # also needs to authorize full access to Amazon EC2, because Cloudbridge
        # leverages EC2-specific functions to assert the credentials.
        #
        # Therefore, to adhere with principle of least privilege, we do not
        # assert credentials; instead, we handle exceptions raised as a
        # result of signing API calls to cloud provider (e.g., GCP) using
        # incorrect, invalid, or unauthorized credentials.

        return connection

    @classmethod
    def parse_xml(clazz, config_xml):
        # The following reads common cloud-based storage configuration
        # as implemented for the S3 backend. Hence, it also attempts to
        # parse S3-specific configuration (e.g., credentials); however,
        # such provider-specific configuration is overwritten in the
        # following.
        config = parse_config_xml(config_xml)

        try:
            provider = config_xml.attrib.get("provider")
            if provider is None:
                msg = "Missing `provider` attribute from the Cloud backend of the ObjectStore."
                log.error(msg)
                raise Exception(msg)
            provider = provider.lower()
            config["provider"] = provider

            # Read any provider-specific configuration.
            auth_element = config_xml.findall("auth")[0]
            missing_config = []
            if provider == "aws":
                akey = auth_element.get("access_key")
                skey = auth_element.get("secret_key")
                config["auth"] = {"access_key": akey, "secret_key": skey}
                if "region" in auth_element:
                    config["auth"]["region"] = auth_element["region"]
            elif provider == "azure":
                sid = auth_element.get("subscription_id")
                if sid is None:
                    missing_config.append("subscription_id")
                cid = auth_element.get("client_id")
                if cid is None:
                    missing_config.append("client_id")
                sec = auth_element.get("secret")
                if sec is None:
                    missing_config.append("secret")
                ten = auth_element.get("tenant")
                if ten is None:
                    missing_config.append("tenant")
                config["auth"] = {"subscription_id": sid, "client_id": cid, "secret": sec, "tenant": ten}
            elif provider == "google":
                cre = auth_element.get("credentials_file")
                if not os.path.isfile(cre):
                    msg = f"The following file specified for GCP credentials not found: {cre}"
                    log.error(msg)
                    raise OSError(msg)
                if cre is None:
                    missing_config.append("credentials_file")
                config["auth"] = {"credentials_file": cre}
            else:
                msg = f"Unsupported provider `{provider}`."
                log.error(msg)
                raise Exception(msg)

            if len(missing_config) > 0:
                msg = (
                    f"The following configuration required for {provider} cloud backend "
                    f"are missing: {missing_config}"
                )
                log.error(msg)
                raise Exception(msg)
            else:
                return config
        except Exception:
            log.exception("Malformed ObjectStore Configuration XML -- unable to continue")
            raise

    def to_dict(self):
        as_dict = super().to_dict()
        as_dict.update(self._config_to_dict())
        return as_dict

    def _config_to_dict(self):
        return {
            "provider": self.provider,
            "auth": self.credentials,
            "bucket": {
                "name": self.bucket_name,
                "use_reduced_redundancy": self.use_rr,
            },
            "cache": {
                "size": self.cache_size,
                "path": self.staging_path,
                "cache_updated_data": self.cache_updated_data,
            },
        }

    def _get_bucket(self, bucket_name):
        try:
            bucket = self.conn.storage.buckets.get(bucket_name)
            if bucket is None:
                log.debug("Bucket not found, creating a bucket with handle '%s'", bucket_name)
                bucket = self.conn.storage.buckets.create(bucket_name)
            log.debug("Using cloud ObjectStore with bucket '%s'", bucket.name)
            return bucket
        except InvalidNameException:
            log.exception("Invalid bucket name -- unable to continue")
            raise
        except Exception:
            # These two generic exceptions will be replaced by specific exceptions
            # once proper exceptions are exposed by CloudBridge.
            log.exception(f"Could not get bucket '{bucket_name}'")
        raise Exception

    def _get_remote_size(self, rel_path):
        try:
            obj = self.bucket.objects.get(rel_path)
            return obj.size
        except Exception:
            log.exception("Could not get size of key '%s' from S3", rel_path)
            return -1

    def _exists_remotely(self, rel_path):
        exists = False
        try:
            # A hackish way of testing if the rel_path is a folder vs a file
            is_dir = rel_path[-1] == "/"
            if is_dir:
                keyresult = self.bucket.objects.list(prefix=rel_path)
                if len(keyresult) > 0:
                    exists = True
                else:
                    exists = False
            else:
                exists = True if self.bucket.objects.get(rel_path) is not None else False
        except Exception:
            log.exception("Trouble checking existence of S3 key '%s'", rel_path)
            return False
        return exists

    def _download(self, rel_path):
        local_destination = self._get_cache_path(rel_path)
        try:
            log.debug("Pulling key '%s' into cache to %s", rel_path, local_destination)
            key = self.bucket.objects.get(rel_path)
            remote_size = key.size
            if not self._caching_allowed(rel_path, remote_size):
                return False
            log.debug("Pulled key '%s' into cache to %s", rel_path, local_destination)
            self._download_to(key, local_destination)
            return True
        except Exception:
            log.exception("Problem downloading key '%s' from S3 bucket '%s'", rel_path, self.bucket.name)
        return False

    def _download_directory_into_cache(self, rel_path, cache_path):
        # List objects in the specified cloud folder
        objects = self.bucket.objects.list(prefix=rel_path)

        for obj in objects:
            remote_file_path = obj.name
            local_file_path = os.path.join(cache_path, os.path.relpath(remote_file_path, rel_path))

            # Create directories if they don't exist
            os.makedirs(os.path.dirname(local_file_path), exist_ok=True)

            # Download the file
            self._download_to(obj, local_file_path)

    def _download_to(self, key, local_destination):
        if self.use_axel:
            url = key.generate_url(7200)
            return self._axel_download(url, local_destination)
        else:
            with open(local_destination, "wb+") as downloaded_file_handle:
                key.save_content(downloaded_file_handle)

    def _push_string_to_path(self, rel_path: str, from_string: str) -> bool:
        try:
            if not self.bucket.objects.get(rel_path):
                created_obj = self.bucket.objects.create(rel_path)
                created_obj.upload(from_string)
            else:
                self.bucket.objects.get(rel_path).upload(from_string)
            return True
        except Exception:
            log.exception("Trouble pushing to cloud '%s' from string", rel_path)
            return False

    def _push_file_to_path(self, rel_path: str, source_file: str) -> bool:
        try:
            if not self.bucket.objects.get(rel_path):
                created_obj = self.bucket.objects.create(rel_path)
                created_obj.upload_from_file(source_file)
            else:
                self.bucket.objects.get(rel_path).upload_from_file(source_file)
            return True
        except Exception:
            log.exception("Trouble pushing to cloud '%s' from file '%s'", rel_path, source_file)
            return False

    def _delete_remote_all(self, rel_path: str) -> bool:
        try:
            results = self.bucket.objects.list(prefix=rel_path)
            for key in results:
                log.debug("Deleting key %s", key.name)
                key.delete()
            return True
        except Exception:
            log.exception("Could not delete key '%s' from cloud", rel_path)
            return False

    def _delete_existing_remote(self, rel_path: str) -> bool:
        try:
            key = self.bucket.objects.get(rel_path)
            log.debug("Deleting key %s", key.name)
            key.delete()
            return True
        except Exception:
            log.exception("Could not delete key '%s' from cloud", rel_path)
            return False

    def _get_object_url(self, obj, **kwargs):
        if self._exists(obj, **kwargs):
            rel_path = self._construct_path(obj, **kwargs)
            try:
                key = self.bucket.objects.get(rel_path)
                return key.generate_url(expires_in=86400)  # 24hrs
            except Exception:
                log.exception("Trouble generating URL for dataset '%s'", rel_path)
        return None

    def _get_store_usage_percent(self, obj):
        return 0.0

    def shutdown(self):
        self._shutdown_cache_monitor()



"""
Object Store plugin for the Amazon Simple Storage Service (S3)
"""

import logging
import os
import time
from datetime import datetime

try:
    # Imports are done this way to allow objectstore code to be used outside of Galaxy.
    import boto
    from boto.exception import S3ResponseError
    from boto.s3.connection import S3Connection
    from boto.s3.key import Key
except ImportError:
    boto = None  # type: ignore[assignment]

from galaxy.util import string_as_bool
from ._caching_base import CachingConcreteObjectStore
from ._util import UsesAxel
from .caching import (
    enable_cache_monitor,
    parse_caching_config_dict_from_xml,
)
from .s3_multipart_upload import multipart_upload

NO_BOTO_ERROR_MESSAGE = (
    "S3/Swift object store configured, but no boto dependency available."
    "Please install and properly configure boto or modify object store configuration."
)

log = logging.getLogger(__name__)
logging.getLogger("boto").setLevel(logging.INFO)  # Otherwise boto is quite noisy


def download_directory(bucket, remote_folder, local_path):
    # List objects in the specified S3 folder
    objects = bucket.list(prefix=remote_folder)

    for obj in objects:
        remote_file_path = obj.key
        local_file_path = os.path.join(local_path, os.path.relpath(remote_file_path, remote_folder))

        # Create directories if they don't exist
        os.makedirs(os.path.dirname(local_file_path), exist_ok=True)

        # Download the file
        obj.get_contents_to_filename(local_file_path)


def parse_config_xml(config_xml):
    try:
        a_xml = config_xml.findall("auth")[0]
        access_key = a_xml.get("access_key")
        secret_key = a_xml.get("secret_key")

        b_xml = config_xml.findall("bucket")[0]
        bucket_name = b_xml.get("name")
        use_rr = string_as_bool(b_xml.get("use_reduced_redundancy", "False"))
        max_chunk_size = int(b_xml.get("max_chunk_size", 250))

        cn_xml = config_xml.findall("connection")
        if not cn_xml:
            cn_xml = {}
        else:
            cn_xml = cn_xml[0]

        host = cn_xml.get("host", None)
        port = int(cn_xml.get("port", 6000))
        multipart = string_as_bool(cn_xml.get("multipart", "True"))
        is_secure = string_as_bool(cn_xml.get("is_secure", "True"))
        conn_path = cn_xml.get("conn_path", "/")
        region = cn_xml.get("region", None)

        cache_dict = parse_caching_config_dict_from_xml(config_xml)

        tag, attrs = "extra_dir", ("type", "path")
        extra_dirs = config_xml.findall(tag)
        if not extra_dirs:
            msg = f"No {tag} element in XML tree"
            log.error(msg)
            raise Exception(msg)
        extra_dirs = [{k: e.get(k) for k in attrs} for e in extra_dirs]

        config_dict = {
            "auth": {
                "access_key": access_key,
                "secret_key": secret_key,
            },
            "bucket": {
                "name": bucket_name,
                "use_reduced_redundancy": use_rr,
                "max_chunk_size": max_chunk_size,
            },
            "connection": {
                "host": host,
                "port": port,
                "multipart": multipart,
                "is_secure": is_secure,
                "conn_path": conn_path,
                "region": region,
            },
            "cache": cache_dict,
            "extra_dirs": extra_dirs,
            "private": CachingConcreteObjectStore.parse_private_from_config_xml(config_xml),
        }
        name = config_xml.attrib.get("name", None)
        if name is not None:
            config_dict["name"] = name
        device = config_xml.attrib.get("device", None)
        config_dict["device"] = device
        return config_dict
    except Exception:
        # Toss it back up after logging, we can't continue loading at this point.
        log.exception("Malformed ObjectStore Configuration XML -- unable to continue")
        raise


class CloudConfigMixin:
    def _config_to_dict(self):
        return {
            "auth": {
                "access_key": self.access_key,
                "secret_key": self.secret_key,
            },
            "bucket": {
                "name": self.bucket,
                "use_reduced_redundancy": self.use_rr,
            },
            "connection": {
                "host": self.host,
                "port": self.port,
                "multipart": self.multipart,
                "is_secure": self.is_secure,
                "conn_path": self.conn_path,
                "region": self.region,
            },
            "cache": {
                "size": self.cache_size,
                "path": self.staging_path,
                "cache_updated_data": self.cache_updated_data,
            },
        }


class S3ObjectStore(CachingConcreteObjectStore, CloudConfigMixin, UsesAxel):
    """
    Object store that stores objects as items in an AWS S3 bucket. A local
    cache exists that is used as an intermediate location for files between
    Galaxy and S3.
    """

    store_type = "aws_s3"
    cloud = True

    def __init__(self, config, config_dict):
        super().__init__(config, config_dict)
        self.cache_monitor = None

        self.transfer_progress = 0

        auth_dict = config_dict["auth"]
        bucket_dict = config_dict["bucket"]
        connection_dict = config_dict.get("connection", {})
        cache_dict = config_dict.get("cache") or {}
        self.enable_cache_monitor, self.cache_monitor_interval = enable_cache_monitor(config, config_dict)

        self.access_key = auth_dict.get("access_key")
        self.secret_key = auth_dict.get("secret_key")

        self.bucket = bucket_dict.get("name")
        self.use_rr = bucket_dict.get("use_reduced_redundancy", False)
        self.max_chunk_size = bucket_dict.get("max_chunk_size", 250)

        self.host = connection_dict.get("host", None)
        self.port = connection_dict.get("port", 6000)
        self.multipart = connection_dict.get("multipart", True)
        self.is_secure = connection_dict.get("is_secure", True)
        self.conn_path = connection_dict.get("conn_path", "/")
        self.region = connection_dict.get("region", None)

        self.cache_size = cache_dict.get("size") or self.config.object_store_cache_size
        self.staging_path = cache_dict.get("path") or self.config.object_store_cache_path
        self.cache_updated_data = cache_dict.get("cache_updated_data", True)

        extra_dirs = {e["type"]: e["path"] for e in config_dict.get("extra_dirs", [])}
        self.extra_dirs.update(extra_dirs)

        self._initialize()

    def _initialize(self):
        if boto is None:
            raise Exception(NO_BOTO_ERROR_MESSAGE)

        # for multipart upload
        self.s3server = {
            "access_key": self.access_key,
            "secret_key": self.secret_key,
            "is_secure": self.is_secure,
            "max_chunk_size": self.max_chunk_size,
            "host": self.host,
            "port": self.port,
            "use_rr": self.use_rr,
            "conn_path": self.conn_path,
        }

        self._ensure_staging_path_writable()
        self._configure_connection()
        self._bucket = self._get_bucket(self.bucket)
        self._start_cache_monitor_if_needed()
        self._init_axel()

    def _configure_connection(self):
        log.debug("Configuring S3 Connection")
        # If access_key is empty use default credential chain
        if self.access_key:
            if self.region:
                # If specify a region we can infer a host and turn on SIGV4.
                # https://stackoverflow.com/questions/26744712/s3-using-boto-and-sigv4-missing-host-parameter

                # Turning on SIGV4 is needed for AWS regions created after 2014... from
                # https://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-authenticating-requests.html:
                #
                # "Amazon S3 supports Signature Version 4, a protocol for authenticating inbound API requests to AWS services,
                # in all AWS Regions. At this time, AWS Regions created before January 30, 2014 will continue to support the
                # previous protocol, Signature Version 2. Any new Regions after January 30, 2014 will support only Signature
                # Version 4 and therefore all requests to those Regions must be made with Signature Version 4."
                os.environ["S3_USE_SIGV4"] = "True"
                self.conn = S3Connection(self.access_key, self.secret_key, host=f"s3.{self.region}.amazonaws.com")
            else:
                # See notes above, this path through the code will not work for
                # newer regions.
                self.conn = S3Connection(self.access_key, self.secret_key)
        else:
            self.conn = S3Connection()

    @classmethod
    def parse_xml(clazz, config_xml):
        return parse_config_xml(config_xml)

    def to_dict(self):
        as_dict = super().to_dict()
        as_dict.update(self._config_to_dict())
        return as_dict

    def _get_bucket(self, bucket_name):
        """Sometimes a handle to a bucket is not established right away so try
        it a few times. Raise error is connection is not established."""
        last_error = None
        for i in range(5):
            try:
                bucket = self.conn.get_bucket(bucket_name)
                log.debug("Using cloud object store with bucket '%s'", bucket.name)
                return bucket
            except S3ResponseError as e:
                last_error = e
                try:
                    log.debug("Bucket not found, creating s3 bucket with handle '%s'", bucket_name)
                    self.conn.create_bucket(bucket_name)
                except S3ResponseError:
                    log.exception("Could not get bucket '%s', attempt %s/5", bucket_name, i + 1)
                    time.sleep(2)
        # All the attempts have been exhausted and connection was not established,
        # raise error
        if last_error:
            raise last_error
        else:
            raise Exception("Failed to connect to target object store.")

    def _get_transfer_progress(self):
        return self.transfer_progress

    def _get_remote_size(self, rel_path):
        try:
            key = self._bucket.get_key(rel_path)
            return key.size
        except (S3ResponseError, AttributeError):
            log.exception("Could not get size of key '%s' from S3", rel_path)
            return -1

    def _exists_remotely(self, rel_path):
        exists = False
        try:
            # A hackish way of testing if the rel_path is a folder vs a file
            is_dir = rel_path[-1] == "/"
            if is_dir:
                keyresult = self._bucket.get_all_keys(prefix=rel_path)
                if len(keyresult) > 0:
                    exists = True
                else:
                    exists = False
            else:
                key = Key(self._bucket, rel_path)
                exists = key.exists()
        except S3ResponseError:
            log.exception("Trouble checking existence of S3 key '%s'", rel_path)
            return False
        return exists

    def _transfer_cb(self, complete, total):
        self.transfer_progress += 10

    def _download(self, rel_path):
        local_destination = self._get_cache_path(rel_path)
        try:
            log.debug("Pulling key '%s' into cache to %s", rel_path, local_destination)
            key = self._bucket.get_key(rel_path)
            if key is None:
                message = f"Attempting to download an invalid key for path {rel_path}."
                log.critical(message)
                raise Exception(message)
            remote_size = key.size
            if not self._caching_allowed(rel_path, remote_size):
                return False
            if self.use_axel:
                log.debug("Parallel pulled key '%s' into cache to %s", rel_path, local_destination)
                url = key.generate_url(7200)
                return self._axel_download(url, local_destination)
            else:
                log.debug("Pulled key '%s' into cache to %s", rel_path, local_destination)
                self.transfer_progress = 0  # Reset transfer progress counter
                key.get_contents_to_filename(local_destination, cb=self._transfer_cb, num_cb=10)
                return True
        except S3ResponseError:
            log.exception("Problem downloading key '%s' from S3 bucket '%s'", rel_path, self._bucket.name)
        return False

    def _push_to_storage(self, rel_path, source_file=None, from_string=None):
        """
        Push the file pointed to by ``rel_path`` to the object store naming the key
        ``rel_path``. If ``source_file`` is provided, push that file instead while
        still using ``rel_path`` as the key name.
        If ``from_string`` is provided, set contents of the file to the value of
        the string.
        """
        try:
            source_file = source_file if source_file else self._get_cache_path(rel_path)
            if os.path.exists(source_file):
                key = Key(self._bucket, rel_path)
                if os.path.getsize(source_file) == 0 and key.exists():
                    log.debug(
                        "Wanted to push file '%s' to S3 key '%s' but its size is 0; skipping.", source_file, rel_path
                    )
                    return True
                if from_string:
                    key.set_contents_from_string(from_string, reduced_redundancy=self.use_rr)
                    log.debug("Pushed data from string '%s' to key '%s'", from_string, rel_path)
                else:
                    start_time = datetime.now()
                    log.debug(
                        "Pushing cache file '%s' of size %s bytes to key '%s'",
                        source_file,
                        os.path.getsize(source_file),
                        rel_path,
                    )
                    mb_size = os.path.getsize(source_file) / 1e6
                    if mb_size < 10 or (not self.multipart):
                        self.transfer_progress = 0  # Reset transfer progress counter
                        key.set_contents_from_filename(
                            source_file, reduced_redundancy=self.use_rr, cb=self._transfer_cb, num_cb=10
                        )
                    else:
                        multipart_upload(self.s3server, self._bucket, key.name, source_file, mb_size)
                    end_time = datetime.now()
                    log.debug(
                        "Pushed cache file '%s' to key '%s' (%s bytes transfered in %s sec)",
                        source_file,
                        rel_path,
                        os.path.getsize(source_file),
                        end_time - start_time,
                    )
                return True
            else:
                log.error(
                    "Tried updating key '%s' from source file '%s', but source file does not exist.",
                    rel_path,
                    source_file,
                )
        except S3ResponseError:
            log.exception("Trouble pushing S3 key '%s' from file '%s'", rel_path, source_file)
            raise
        return False

    def _delete_remote_all(self, rel_path: str) -> bool:
        try:
            results = self._bucket.get_all_keys(prefix=rel_path)
            for key in results:
                log.debug("Deleting key %s", key.name)
                key.delete()
            return True
        except S3ResponseError:
            log.exception("Could not delete blob '%s' from S3", rel_path)
            return False

    def _delete_existing_remote(self, rel_path: str) -> bool:
        try:
            key = Key(self._bucket, rel_path)
            log.debug("Deleting key %s", key.name)
            key.delete()
            return True
        except S3ResponseError:
            log.exception("Could not delete blob '%s' from S3", rel_path)
            return False

    def _download_directory_into_cache(self, rel_path, cache_path):
        download_directory(self._bucket, rel_path, cache_path)

    def _get_object_url(self, obj, **kwargs):
        if self._exists(obj, **kwargs):
            rel_path = self._construct_path(obj, **kwargs)
            try:
                key = Key(self._bucket, rel_path)
                return key.generate_url(expires_in=86400)  # 24hrs
            except S3ResponseError:
                log.exception("Trouble generating URL for dataset '%s'", rel_path)
        return None

    def _get_store_usage_percent(self, obj):
        return 0.0

    def shutdown(self):
        self._shutdown_cache_monitor()


class GenericS3ObjectStore(S3ObjectStore):
    """
    Object store that stores objects as items in a generic S3 (non AWS) bucket.
    A local cache exists that is used as an intermediate location for files between
    Galaxy and the S3 storage service.
    """

    store_type = "generic_s3"

    def _configure_connection(self):
        log.debug("Configuring generic S3 Connection")
        self.conn = boto.connect_s3(
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            is_secure=self.is_secure,
            host=self.host,
            port=self.port,
            calling_format=boto.s3.connection.OrdinaryCallingFormat(),
            path=self.conn_path,
        )



"""
objectstore package, abstraction for storing blobs of data for use in Galaxy.

all providers ensure that data can be accessed on the filesystem for running
tools
"""

import abc
import logging
import os
import random
import shutil
import threading
import time
from typing import (
    Any,
    Dict,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    TYPE_CHECKING,
)
from uuid import uuid4

import yaml
from pydantic import BaseModel
from typing_extensions import (
    Literal,
    Protocol,
)

from galaxy.exceptions import (
    ConfigDoesNotAllowException,
    MalformedContents,
    ObjectInvalid,
    ObjectNotFound,
)
from galaxy.util import (
    asbool,
    directory_hash_id,
    force_symlink,
    in_directory,
    parse_xml,
    umask_fix_perms,
)
from galaxy.util.bunch import Bunch
from galaxy.util.path import (
    safe_makedirs,
    safe_relpath,
    safe_walk,
)
from galaxy.util.sleeper import Sleeper
from .badges import (
    BadgeDict,
    read_badges,
    serialize_badges,
    StoredBadgeDict,
)
from .caching import CacheTarget
from .templates import ObjectStoreConfiguration

if TYPE_CHECKING:
    from galaxy.model import (
        Dataset,
        DatasetInstance,
    )

NO_SESSION_ERROR_MESSAGE = (
    "Attempted to 'create' object store entity in configuration with no database session present."
)
DEFAULT_PRIVATE = False
DEFAULT_QUOTA_SOURCE = None  # Just track quota right on user object in Galaxy.
DEFAULT_QUOTA_ENABLED = True  # enable quota tracking in object stores by default
DEFAULT_DEVICE_ID = None
log = logging.getLogger(__name__)


class UserObjectStoreResolver(Protocol):
    def resolve_object_store_uri_config(self, uri: str) -> ObjectStoreConfiguration:
        pass

    def resolve_object_store_uri(self, uri: str) -> "ConcreteObjectStore":
        pass


class BaseUserObjectStoreResolver(UserObjectStoreResolver, metaclass=abc.ABCMeta):
    _app_config: "UserObjectStoresAppConfig"

    @abc.abstractmethod
    def resolve_object_store_uri_config(self, uri: str) -> ObjectStoreConfiguration:
        """Resolve the supplied object store URI into a concrete object store configuration."""
        pass

    def resolve_object_store_uri(self, uri: str) -> "ConcreteObjectStore":
        object_store_configuration = self.resolve_object_store_uri_config(uri)
        return concrete_object_store(object_store_configuration, self._app_config)


class ObjectStore(metaclass=abc.ABCMeta):
    """ObjectStore interface.

    FIELD DESCRIPTIONS (these apply to all the methods in this class):

    :type obj: StorableObject
    :param obj: A Galaxy object with an assigned database ID accessible via
        the .id attribute.

    :type base_dir: string
    :param base_dir: A key in `self.extra_dirs` corresponding to the base
        directory in which this object should be created, or `None` to specify
        the default directory.

    :type dir_only: boolean
    :param dir_only: If `True`, check only the path where the file identified
        by `obj` should be located, not the dataset itself. This option applies
        to `extra_dir` argument as well.

    :type extra_dir: string
    :param extra_dir: Append `extra_dir` to the directory structure where the
        dataset identified by `obj` should be located. (e.g.,
        000/extra_dir/obj.id). Valid values include 'job_work' (defaulting to
        config.jobs_directory =
        '$GALAXY_ROOT/database/jobs_directory');
        'temp' (defaulting to config.new_file_path =
        '$GALAXY_ROOT/database/tmp').

    :type extra_dir_at_root: boolean
    :param extra_dir_at_root: Applicable only if `extra_dir` is set. If True,
        the `extra_dir` argument is placed at root of the created directory
        structure rather than at the end (e.g., extra_dir/000/obj.id vs.
        000/extra_dir/obj.id)

    :type alt_name: string
    :param alt_name: Use this name as the alternative name for the created
        dataset rather than the default.

    :type obj_dir: boolean
    :param obj_dir: Append a subdirectory named with the object's ID (e.g.
        000/obj.id)
    """

    @abc.abstractmethod
    def exists(self, obj, base_dir=None, dir_only=False, extra_dir=None, extra_dir_at_root=False, alt_name=None):
        """Return True if the object identified by `obj` exists, False otherwise."""
        raise NotImplementedError()

    @abc.abstractmethod
    def construct_path(
        self, obj, base_dir=None, dir_only=False, extra_dir=None, extra_dir_at_root=False, alt_name=None
    ):
        raise NotImplementedError()

    @abc.abstractmethod
    def create(
        self, obj, base_dir=None, dir_only=False, extra_dir=None, extra_dir_at_root=False, alt_name=None, obj_dir=False
    ):
        """
        Mark the object (`obj`) as existing in the store, but with no content.

        This method will create a proper directory structure for
        the file if the directory does not already exist.

        The method returns the concrete objectstore the supplied object is stored
        in.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def empty(self, obj, base_dir=None, extra_dir=None, extra_dir_at_root=False, alt_name=None, obj_dir=False):
        """
        Test if the object identified by `obj` has content.

        If the object does not exist raises `ObjectNotFound`.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def size(self, obj, extra_dir=None, extra_dir_at_root=False, alt_name=None, obj_dir=False) -> int:
        """
        Return size of the object identified by `obj`.

        If the object does not exist, return 0.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def delete(
        self,
        obj,
        entire_dir=False,
        base_dir=None,
        extra_dir=None,
        extra_dir_at_root=False,
        alt_name=None,
        obj_dir=False,
    ):
        """
        Delete the object identified by `obj`.

        :type entire_dir: boolean
        :param entire_dir: If True, delete the entire directory pointed to by
                           extra_dir. For safety reasons, this option applies
                           only for and in conjunction with the extra_dir or
                           obj_dir options.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_data(
        self,
        obj,
        start=0,
        count=-1,
        base_dir=None,
        extra_dir=None,
        extra_dir_at_root=False,
        alt_name=None,
        obj_dir=False,
    ):
        """
        Fetch `count` bytes of data offset by `start` bytes using `obj.id`.

        If the object does not exist raises `ObjectNotFound`.

        :type start: int
        :param start: Set the position to start reading the dataset file

        :type count: int
        :param count: Read at most `count` bytes from the dataset
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_filename(
        self, obj, base_dir=None, dir_only=False, extra_dir=None, extra_dir_at_root=False, alt_name=None, obj_dir=False
    ):
        """
        Get the expected filename with absolute path for object with id `obj.id`.

        This can be used to access the contents of the object.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def update_from_file(
        self,
        obj,
        base_dir=None,
        extra_dir=None,
        extra_dir_at_root=False,
        alt_name=None,
        obj_dir=False,
        file_name=None,
        create=False,
        preserve_symlinks=False,
    ):
        """
        Inform the store that the file associated with `obj.id` has been updated.

        If `file_name` is provided, update from that file instead of the
        default.
        If the object does not exist raises `ObjectNotFound`.

        :type file_name: string
        :param file_name: Use file pointed to by `file_name` as the source for
                          updating the dataset identified by `obj`

        :type create: boolean
        :param create: If True and the default dataset does not exist, create
            it first.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_object_url(self, obj, extra_dir=None, extra_dir_at_root=False, alt_name=None, obj_dir=False):
        """
        Return the URL for direct access if supported, otherwise return None.

        Note: need to be careful to not bypass dataset security with this.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_concrete_store_name(self, obj):
        """Return a display name or title of the objectstore corresponding to obj.

        To accommodate nested objectstores, obj is passed in so this metadata can
        be returned for the ConcreteObjectStore corresponding to the object.

        If the dataset is in a new or discarded state and an object_store_id has not
        yet been set, this may return ``None``.
        """

    @abc.abstractmethod
    def get_concrete_store_description_markdown(self, obj):
        """Return a longer description of how data 'obj' is stored.

        To accommodate nested objectstores, obj is passed in so this metadata can
        be returned for the ConcreteObjectStore corresponding to the object.

        If the dataset is in a new or discarded state and an object_store_id has not
        yet been set, this may return ``None``.
        """

    @abc.abstractmethod
    def get_concrete_store_badges(self, obj) -> List[BadgeDict]:
        """Return a list of dictified badges summarizing the object store configuration."""

    @abc.abstractmethod
    def is_private(self, obj):
        """Return True iff supplied object is stored in private ConcreteObjectStore."""

    def object_store_ids(self, private=None):
        """Return IDs of all concrete object stores - either private ones or non-private ones.

        This should just return an empty list for non-DistributedObjectStore object stores,
        i.e. concrete objectstores and the HierarchicalObjectStore since these do not
        use the object_store_id column for objects (Galaxy Datasets).
        """
        return []

    def object_store_allows_id_selection(self) -> bool:
        """Return True if this object store respects object_store_id and allow selection of this."""
        return False

    def validate_selected_object_store_id(self, user, object_store_id: Optional[str]) -> Optional[str]:
        if object_store_id and not self.object_store_allows_id_selection():
            return "The current configuration doesn't allow selecting preferred object stores."
        return None

    def object_store_ids_allowing_selection(self) -> List[str]:
        """Return a non-emtpy list of allowed selectable object store IDs during creation."""
        return []

    def get_concrete_store_by_object_store_id(self, object_store_id: str) -> Optional["ConcreteObjectStore"]:
        """If this is a distributed object store, get ConcreteObjectStore by object_store_id."""
        return None

    @abc.abstractmethod
    def get_store_usage_percent(self):
        """Return the percentage indicating how full the store is."""
        raise NotImplementedError()

    @abc.abstractmethod
    def get_store_by(self, obj):
        """Return how object is stored (by 'uuid', 'id', or None if not yet saved).

        Certain Galaxy remote data features aren't available if objects are stored by 'id'.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def cache_targets(self) -> List[CacheTarget]:
        """Return a list of CacheTargets used by this object store."""
        raise NotImplementedError()

    @abc.abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_quota_source_map(self):
        """Return QuotaSourceMap describing mapping of object store IDs to quota sources."""

    @abc.abstractmethod
    def get_device_source_map(self) -> "DeviceSourceMap":
        """Return DeviceSourceMap describing mapping of object store IDs to device sources."""


class BaseObjectStore(ObjectStore):
    store_by: str
    store_type: str

    def __init__(self, config, config_dict=None, **kwargs):
        """
        :type config: object
        :param config: An object, most likely populated from
            `galaxy/config.ini`, having the following attributes:

            * object_store_check_old_style (only used by the
              :class:`DiskObjectStore` subclass)
            * jobs_directory -- Each job is given a unique empty directory
              as its current working directory. This option defines in what
              parent directory those directories will be created.
            * new_file_path -- Used to set the 'temp' extra_dir.
        """
        if config_dict is None:
            config_dict = {}
        self.running = True
        self.config = config
        self.check_old_style = config.object_store_check_old_style
        self.galaxy_enable_quotas = config.enable_quotas
        extra_dirs = {}
        extra_dirs["job_work"] = config.jobs_directory
        extra_dirs["temp"] = config.new_file_path
        extra_dirs.update({e["type"]: e["path"] for e in config_dict.get("extra_dirs", [])})
        self.extra_dirs = extra_dirs

    def start(self):
        """
        Call all postfork function(s) here. These functions spawn a thread.
        We register start(self) with app, so app starts the threads. Override
        this function in subclasses, as needed.
        """

    def shutdown(self):
        """Close any connections for this ObjectStore."""
        self.running = False

    @classmethod
    def parse_xml(clazz, config_xml):
        """Parse an XML description of a configuration for this object store.

        Return a configuration dictionary (such as would correspond to the YAML configuration)
        for the object store.
        """
        raise NotImplementedError()

    @classmethod
    def from_xml(clazz, config, config_xml, **kwd):
        config_dict = clazz.parse_xml(config_xml)
        return clazz(config, config_dict, **kwd)

    def to_dict(self):
        extra_dirs = []
        for extra_dir_type, extra_dir_path in self.extra_dirs.items():
            extra_dirs.append({"type": extra_dir_type, "path": extra_dir_path})
        store_type = self.store_type
        return {
            "config": config_to_dict(self.config),
            "extra_dirs": extra_dirs,
            "type": store_type,
        }

    def _get_object_id(self, obj):
        if hasattr(obj, self.store_by):
            obj_id = getattr(obj, self.store_by)
            if obj_id is None:
                obj.flush()
                return obj.id
            return obj_id
        else:
            # job's don't have uuids, so always use ID in this case when creating
            # job working directories.
            return obj.id

    def _invoke(self, delegate, obj=None, **kwargs):
        return self.__getattribute__(f"_{delegate}")(obj=obj, **kwargs)

    def exists(self, obj, **kwargs):
        return self._invoke("exists", obj, **kwargs)

    def construct_path(self, obj, **kwargs):
        return self._invoke("construct_path", obj, **kwargs)

    def create(self, obj, **kwargs):
        return self._invoke("create", obj, **kwargs)

    def empty(self, obj, **kwargs):
        return self._invoke("empty", obj, **kwargs)

    def size(self, obj, **kwargs):
        return self._invoke("size", obj, **kwargs)

    def delete(self, obj, **kwargs):
        return self._invoke("delete", obj, **kwargs)

    def get_data(self, obj, **kwargs):
        return self._invoke("get_data", obj, **kwargs)

    def get_filename(self, obj, **kwargs):
        return self._invoke("get_filename", obj, **kwargs)

    def update_from_file(self, obj, **kwargs):
        return self._invoke("update_from_file", obj, **kwargs)

    def get_object_url(self, obj, **kwargs):
        return self._invoke("get_object_url", obj, **kwargs)

    def get_concrete_store_name(self, obj):
        return self._invoke("get_concrete_store_name", obj)

    def get_concrete_store_description_markdown(self, obj):
        return self._invoke("get_concrete_store_description_markdown", obj)

    def get_concrete_store_badges(self, obj) -> List[BadgeDict]:
        return self._invoke("get_concrete_store_badges", obj)

    def get_store_usage_percent(self):
        return self._invoke("get_store_usage_percent")

    def get_store_by(self, obj, **kwargs):
        return self._invoke("get_store_by", obj, **kwargs)

    def is_private(self, obj):
        return self._invoke("is_private", obj)

    def cache_targets(self) -> List[CacheTarget]:
        return []

    @classmethod
    def parse_private_from_config_xml(clazz, config_xml):
        private = DEFAULT_PRIVATE
        if config_xml is not None:
            private = asbool(config_xml.attrib.get("private", DEFAULT_PRIVATE))
        return private

    @classmethod
    def parse_badges_from_config_xml(clazz, badges_xml):
        badges = []
        for e in badges_xml:
            type = e.tag
            message = e.text
            badges.append({"type": type, "message": message})
        return badges

    def get_quota_source_map(self):
        # I'd rather keep this abstract... but register_singleton wants it to be instantiable...
        raise NotImplementedError()

    def get_device_source_map(self):
        return DeviceSourceMap()


class ConcreteObjectStore(BaseObjectStore):
    """Subclass of ObjectStore for stores that don't delegate (non-nested).

    Adds store_by and quota_source functionality. These attributes do not make
    sense for the delegating object stores, they should describe files at actually
    persisted, not how a file is routed to a persistence source.
    """

    badges: List[StoredBadgeDict]
    device_id: Optional[str] = None
    cloud: bool = False

    def __init__(self, config, config_dict=None, **kwargs):
        """
        :type config: object
        :param config: An object, most likely populated from
            `galaxy/config.ini`, having the following attributes:

            * object_store_check_old_style (only used by the
              :class:`DiskObjectStore` subclass)
            * jobs_directory -- Each job is given a unique empty directory
              as its current working directory. This option defines in what
              parent directory those directories will be created.
            * new_file_path -- Used to set the 'temp' extra_dir.
        """
        if config_dict is None:
            config_dict = {}
        super().__init__(config=config, config_dict=config_dict, **kwargs)
        self.store_by = config_dict.get("store_by", None) or getattr(config, "object_store_store_by", "id")
        self.name = config_dict.get("name", None)
        self.description = config_dict.get("description", None)
        # Annotate this as true to prevent sharing of data.
        self.private = config_dict.get("private", DEFAULT_PRIVATE)
        # short label describing the quota source or null to use default
        # quota source right on user object.
        quota_config = config_dict.get("quota", {})
        self.quota_source = quota_config.get("source", DEFAULT_QUOTA_SOURCE)
        self.quota_enabled = quota_config.get("enabled", DEFAULT_QUOTA_ENABLED)
        self.device_id = config_dict.get("device", None)
        self.badges = read_badges(config_dict)

    def to_dict(self):
        rval = super().to_dict()
        rval["private"] = self.private
        rval["store_by"] = self.store_by
        rval["name"] = self.name
        rval["description"] = self.description
        rval["quota"] = {
            "source": self.quota_source,
            "enabled": self.quota_enabled,
        }
        rval["badges"] = self._get_concrete_store_badges(None)
        rval["device"] = self.device_id
        return rval

    def to_model(self, object_store_id: str) -> "ConcreteObjectStoreModel":
        return ConcreteObjectStoreModel(
            object_store_id=object_store_id,
            private=self.private,
            name=self.name,
            description=self.description,
            quota=QuotaModel(source=self.quota_source, enabled=self.quota_enabled),
            badges=self._get_concrete_store_badges(None),
            device=self.device_id,
        )

    def _get_concrete_store_badges(self, obj) -> List[BadgeDict]:
        return serialize_badges(
            self.badges,
            self.galaxy_enable_quotas and self.quota_enabled,
            self.private,
            False,
            self.cloud,
        )

    def _get_concrete_store_name(self, obj):
        return self.name

    def _get_concrete_store_description_markdown(self, obj):
        return self.description

    def _get_store_by(self, obj):
        return self.store_by

    def _is_private(self, obj):
        return self.private

    @property
    def cache_target(self) -> Optional[CacheTarget]:
        return None

    def cache_targets(self) -> List[CacheTarget]:
        cache_target = self.cache_target
        return [cache_target] if cache_target is not None else []

    def get_quota_source_map(self):
        quota_source_map = QuotaSourceMap(
            self.quota_source,
            self.quota_enabled,
        )
        return quota_source_map

    def get_device_source_map(self) -> "DeviceSourceMap":
        return DeviceSourceMap(self.device_id)


class DiskObjectStore(ConcreteObjectStore):
    """
    Standard Galaxy object store.

    Stores objects in files under a specific directory on disk.

    >>> from galaxy.util.bunch import Bunch
    >>> import tempfile
    >>> file_path=tempfile.mkdtemp()
    >>> obj = Bunch(id=1)
    >>> s = DiskObjectStore(Bunch(umask=0o077, jobs_directory=file_path, new_file_path=file_path, object_store_check_old_style=False, enable_quotas=True), dict(files_dir=file_path))
    >>> o = s.create(obj)
    >>> s.exists(obj)
    True
    >>> assert s.get_filename(obj) == file_path + '/000/dataset_1.dat'
    """

    store_type = "disk"

    def __init__(self, config, config_dict):
        """
        :type config: object
        :param config: An object, most likely populated from
            `galaxy/config.ini`, having the same attributes needed by
            :class:`ObjectStore` plus:

            * file_path -- Default directory to store objects to disk in.
            * umask -- the permission bits for newly created files.

        :type file_path: str
        :param file_path: Override for the `config.file_path` value.

        :type extra_dirs: dict
        :param extra_dirs: Keys are string, values are directory paths.
        """
        super().__init__(config, config_dict)
        self.file_path = os.path.abspath(config_dict.get("files_dir") or config.file_path)

    @classmethod
    def parse_xml(clazz, config_xml):
        extra_dirs = []
        config_dict = {}
        if config_xml is not None:
            store_by = config_xml.attrib.get("store_by", None)
            if store_by is not None:
                config_dict["store_by"] = store_by
            name = config_xml.attrib.get("name", None)
            if name is not None:
                config_dict["name"] = name
            device = config_xml.attrib.get("device", None)
            config_dict["device"] = device
            for e in config_xml:
                if e.tag == "quota":
                    config_dict["quota"] = {
                        "source": e.get("source", DEFAULT_QUOTA_SOURCE),
                        "enabled": asbool(e.get("enabled", DEFAULT_QUOTA_ENABLED)),
                    }
                elif e.tag == "files_dir":
                    config_dict["files_dir"] = e.get("path")
                elif e.tag == "description":
                    config_dict["description"] = e.text
                elif e.tag == "badges":
                    config_dict["badges"] = BaseObjectStore.parse_badges_from_config_xml(e)
                else:
                    extra_dirs.append({"type": e.get("type"), "path": e.get("path")})

        config_dict["extra_dirs"] = extra_dirs
        config_dict["private"] = BaseObjectStore.parse_private_from_config_xml(config_xml)
        return config_dict

    def to_dict(self):
        as_dict = super().to_dict()
        as_dict["files_dir"] = self.file_path
        return as_dict

    def __get_filename(
        self, obj, base_dir=None, dir_only=False, extra_dir=None, extra_dir_at_root=False, alt_name=None, obj_dir=False
    ):
        """
        Return the absolute path for the file corresponding to the `obj.id`.

        This is regardless of whether or not the file exists.
        """
        path = self._construct_path(
            obj,
            base_dir=base_dir,
            dir_only=dir_only,
            extra_dir=extra_dir,
            extra_dir_at_root=extra_dir_at_root,
            alt_name=alt_name,
            obj_dir=False,
            old_style=True,
        )
        # For backward compatibility: check the old style root path first;
        # otherwise construct hashed path.
        if not os.path.exists(path):
            return self._construct_path(
                obj,
                base_dir=base_dir,
                dir_only=dir_only,
                extra_dir=extra_dir,
                extra_dir_at_root=extra_dir_at_root,
                alt_name=alt_name,
            )

    # TODO: rename to _disk_path or something like that to avoid conflicts with
    # children that'll use the local_extra_dirs decorator, e.g. S3
    def _construct_path(
        self,
        obj,
        old_style=False,
        base_dir=None,
        dir_only=False,
        extra_dir=None,
        extra_dir_at_root=False,
        alt_name=None,
        obj_dir=False,
        **kwargs,
    ):
        """
        Construct the absolute path for accessing the object identified by `obj.id`.

        :type base_dir: string
        :param base_dir: A key in self.extra_dirs corresponding to the base
                         directory in which this object should be created, or
                         None to specify the default directory.

        :type dir_only: boolean
        :param dir_only: If True, check only the path where the file
                         identified by `obj` should be located, not the
                         dataset itself. This option applies to `extra_dir`
                         argument as well.

        :type extra_dir: string
        :param extra_dir: Append the value of this parameter to the expected
            path used to access the object identified by `obj` (e.g.,
            /files/000/<extra_dir>/dataset_10.dat).

        :type alt_name: string
        :param alt_name: Use this name as the alternative name for the returned
                         dataset rather than the default.

        :type old_style: boolean
        param old_style: This option is used for backward compatibility. If
            `True` then the composed directory structure does not include a
            hash id (e.g., /files/dataset_10.dat (old) vs.
            /files/000/dataset_10.dat (new))
        """
        base = os.path.abspath(self.extra_dirs.get(base_dir) or self.file_path)
        # extra_dir should never be constructed from provided data but just
        # make sure there are no shenannigans afoot
        if extra_dir and extra_dir != os.path.normpath(extra_dir):
            log.warning("extra_dir is not normalized: %s", extra_dir)
            raise ObjectInvalid("The requested object is invalid")
        # ensure that any parent directory references in alt_name would not
        # result in a path not contained in the directory path constructed here
        if alt_name and not safe_relpath(alt_name):
            log.warning("alt_name would locate path outside dir: %s", alt_name)
            raise ObjectInvalid("The requested object is invalid")
        obj_id = self._get_object_id(obj)
        if old_style:
            if extra_dir is not None:
                path = os.path.join(base, extra_dir)
            else:
                path = base
        else:
            # Construct hashed path
            rel_path = os.path.join(*directory_hash_id(obj_id))
            # Create a subdirectory for the object ID
            if obj_dir:
                rel_path = os.path.join(rel_path, str(obj_id))
            # Optionally append extra_dir
            if extra_dir is not None:
                if extra_dir_at_root:
                    rel_path = os.path.join(extra_dir, rel_path)
                else:
                    rel_path = os.path.join(rel_path, extra_dir)
            path = os.path.join(base, rel_path)
        if not dir_only:
            assert (
                obj_id is not None
            ), f"The effective dataset identifier consumed by object store [{self.store_by}] must be set before a path can be constructed."
            path = os.path.join(path, alt_name if alt_name else f"dataset_{obj_id}.dat")
        return os.path.abspath(path)

    def _exists(self, obj, **kwargs):
        """Override `ObjectStore`'s stub and check on disk."""
        if self.check_old_style:
            path = self._construct_path(obj, old_style=True, **kwargs)
            # For backward compatibility: check root path first; otherwise
            # construct and check hashed path.
            if os.path.exists(path):
                return True
        path = self._construct_path(obj, **kwargs)
        return os.path.exists(path)

    def _create(self, obj, **kwargs):
        """Override `ObjectStore`'s stub by creating any files and folders on disk."""
        if not self._exists(obj, **kwargs):
            path = self._construct_path(obj, **kwargs)
            dir_only = kwargs.get("dir_only", False)
            # Create directory if it does not exist
            dir = path if dir_only else os.path.dirname(path)
            safe_makedirs(dir)
            # Create the file if it does not exist
            if not dir_only:
                open(path, "w").close()  # Should be rb?
                umask_fix_perms(path, self.config.umask, 0o666)
        return self

    def _empty(self, obj, **kwargs):
        """Override `ObjectStore`'s stub by checking file size on disk."""
        return self.size(obj, **kwargs) == 0

    def _size(self, obj, **kwargs) -> int:
        """Override `ObjectStore`'s stub by return file size on disk.

        Returns 0 if the object doesn't exist yet or other error.
        """
        if self._exists(obj, **kwargs):
            try:
                filepath = self._get_filename(obj, **kwargs)
                for _ in range(0, 2):
                    size = os.path.getsize(filepath)
                    if size != 0:
                        break
                    # May be legitimately 0, or there may be an issue with the FS / kernel, so we try again
                    time.sleep(0.01)
                return size
            except OSError:
                return 0
        else:
            return 0

    def _delete(self, obj, entire_dir=False, **kwargs):
        """Override `ObjectStore`'s stub; delete the file or folder on disk."""
        path = self._get_filename(obj, **kwargs)
        extra_dir = kwargs.get("extra_dir", None)
        obj_dir = kwargs.get("obj_dir", False)
        try:
            if entire_dir and (extra_dir or obj_dir):
                shutil.rmtree(path)
                return True
            os.remove(path)
            return True
        except FileNotFoundError:
            # Absolutely possible that a delete request races, but that's "fine".
            return True
        except OSError as ex:
            # Likely a race condition in which we delete the job working directory
            # and another process writes files into that directory.
            # If the path doesn't exist anymore, another rmtree call was successful.
            path = self.__get_filename(obj, **kwargs)
            if path is None:
                return True
            else:
                log.critical(f"{path} delete error {ex}", exc_info=True)
        return False

    def _get_data(self, obj, start=0, count=-1, **kwargs):
        """Override `ObjectStore`'s stub; retrieve data directly from disk."""
        data_file = open(self._get_filename(obj, **kwargs))  # Should be rb?
        data_file.seek(start)
        content = data_file.read(count)
        data_file.close()
        return content

    def _get_filename(self, obj, **kwargs):
        """
        Override `ObjectStore`'s stub.

        If `object_store_check_old_style` is set to `True` in config then the
        root path is checked first.
        """
        if self.check_old_style:
            path = self._construct_path(obj, old_style=True, **kwargs)
            # For backward compatibility, check root path first; otherwise,
            # construct and return hashed path
            if os.path.exists(path):
                return path
        path = self._construct_path(obj, **kwargs)
        if not os.path.exists(path):
            raise ObjectNotFound
        return path

    def _update_from_file(self, obj, file_name=None, create=False, **kwargs):
        """`create` parameter is not used in this implementation."""
        preserve_symlinks = kwargs.pop("preserve_symlinks", False)
        # FIXME: symlinks and the object store model may not play well together
        # these should be handled better, e.g. registering the symlink'd file
        # as an object
        if create:
            self._create(obj, **kwargs)
        if file_name and self._exists(obj, **kwargs):
            try:
                if preserve_symlinks and os.path.islink(file_name):
                    force_symlink(os.readlink(file_name), self._get_filename(obj, **kwargs))
                else:
                    path = self._get_filename(obj, **kwargs)
                    shutil.copy(file_name, path)
                    umask_fix_perms(path, self.config.umask, 0o666)
            except shutil.SameFileError:
                # That's ok, we need to ignore this so that remote object stores can update
                # the remote object from the cache file path
                pass
            except OSError as ex:
                log.critical(f"Error copying {file_name} to {self.__get_filename(obj, **kwargs)}: {ex}")
                raise ex

    def _get_object_url(self, obj, **kwargs):
        """
        Override `ObjectStore`'s stub.

        Returns `None`, we have no URLs.
        """
        return None

    def _get_store_usage_percent(self, **kwargs):
        """Override `ObjectStore`'s stub by return percent storage used."""
        st = os.statvfs(self.file_path)
        return (float(st.f_blocks - st.f_bavail) / st.f_blocks) * 100


class NestedObjectStore(BaseObjectStore):
    """
    Base for ObjectStores that use other ObjectStores.

    Example: DistributedObjectStore, HierarchicalObjectStore
    """

    def __init__(self, config, config_xml=None):
        """Extend `ObjectStore`'s constructor."""
        super().__init__(config)
        self.backends = {}

    def shutdown(self):
        """For each backend, shuts them down."""
        for store in self.backends.values():
            store.shutdown()
        super().shutdown()

    def _exists(self, obj, **kwargs):
        """Determine if the `obj` exists in any of the backends."""
        return self._call_method("_exists", obj, False, False, **kwargs)

    def _create(self, obj, **kwargs):
        """Create a backing file in a random backend."""
        objectstore = random.choice(list(self.backends.values()))
        return objectstore.create(obj, **kwargs)

    def cache_targets(self) -> List[CacheTarget]:
        cache_targets = []
        for backend in self.backends.values():
            cache_targets.extend(backend.cache_targets())
        # TODO: merge more intelligently - de-duplicate paths and handle conflicting sizes/percents
        return cache_targets

    def _empty(self, obj, **kwargs):
        """For the first backend that has this `obj`, determine if it is empty."""
        return self._call_method("_empty", obj, True, False, **kwargs)

    def _size(self, obj, **kwargs):
        """For the first backend that has this `obj`, return its size."""
        return self._call_method("_size", obj, 0, False, **kwargs)

    def _delete(self, obj, **kwargs):
        """For the first backend that has this `obj`, delete it."""
        return self._call_method("_delete", obj, False, False, **kwargs)

    def _get_data(self, obj, **kwargs):
        """For the first backend that has this `obj`, get data from it."""
        return self._call_method("_get_data", obj, ObjectNotFound, True, **kwargs)

    def _get_filename(self, obj, **kwargs):
        """For the first backend that has this `obj`, get its filename."""
        return self._call_method("_get_filename", obj, ObjectNotFound, True, **kwargs)

    def _update_from_file(self, obj, **kwargs):
        """For the first backend that has this `obj`, update it from the given file."""
        if kwargs.get("create", False):
            self._create(obj, **kwargs)
            kwargs["create"] = False
        return self._call_method("_update_from_file", obj, ObjectNotFound, True, **kwargs)

    def _get_object_url(self, obj, **kwargs):
        """For the first backend that has this `obj`, get its URL."""
        return self._call_method("_get_object_url", obj, None, False, **kwargs)

    def _get_concrete_store_name(self, obj):
        return self._call_method("_get_concrete_store_name", obj, None, False)

    def _get_concrete_store_description_markdown(self, obj):
        return self._call_method("_get_concrete_store_description_markdown", obj, None, False)

    def _get_concrete_store_badges(self, obj) -> List[BadgeDict]:
        return self._call_method("_get_concrete_store_badges", obj, [], False)

    def _is_private(self, obj):
        return self._call_method("_is_private", obj, False, False)

    def _get_store_by(self, obj):
        return self._call_method("_get_store_by", obj, None, False)

    def _repr_object_for_exception(self, obj):
        try:
            # there are a few objects in python that don't have __class__
            obj_id = self._get_object_id(obj)
            return f"{obj.__class__.__name__}({self.store_by}={obj_id})"
        except AttributeError:
            return str(obj)

    def _call_method(self, method, obj, default, default_is_exception, **kwargs):
        """Check all children object stores for the first one with the dataset."""
        for store in self.backends.values():
            if store.exists(obj, **kwargs):
                return store.__getattribute__(method)(obj, **kwargs)
        if default_is_exception:
            raise default(
                f"objectstore, _call_method failed: {method} on {self._repr_object_for_exception(obj)}, kwargs: {kwargs}"
            )
        else:
            return default


def user_object_store_configuration_to_config_dict(object_store_config: ObjectStoreConfiguration, id) -> Dict[str, Any]:
    # convert a pydantic model describing a user object store into a config dict ready to be
    # slotted into a distributed job runner or stand alone.
    dynamic_object_store_as_dict = object_store_config.model_dump()
    dynamic_object_store_as_dict["id"] = id
    dynamic_object_store_as_dict["weight"] = 0
    # these are all forward facing object stores...
    dynamic_object_store_as_dict["store_by"] = "uuid"
    return dynamic_object_store_as_dict


class DistributedObjectStore(NestedObjectStore):
    """
    ObjectStore that defers to a list of backends.

    When getting objects the first store where the object exists is used.
    When creating objects they are created in a store selected randomly, but
    with weighting.
    """

    store_type = "distributed"
    _quota_source_map: Optional["QuotaSourceMap"]
    _device_source_map: Optional["DeviceSourceMap"]

    def __init__(
        self, config, config_dict, fsmon=False, user_object_store_resolver: Optional[UserObjectStoreResolver] = None
    ):
        """
        :type config: object
        :param config: An object, most likely populated from
            `galaxy/config.ini`, having the same attributes needed by
            :class:`NestedObjectStore` plus:

            * distributed_object_store_config_file

        :type config_xml: ElementTree

        :type fsmon: bool
        :param fsmon: If True, monitor the file system for free space,
            removing backends when they get too full.
        """
        super().__init__(config, config_dict)
        self._quota_source_map = None
        self._device_source_map = None
        self.backends = {}
        self.weighted_backend_ids = []
        self.original_weighted_backend_ids = []
        self.max_percent_full = {}
        self.global_max_percent_full = config_dict.get("global_max_percent_full", 0)
        self.search_for_missing = config_dict.get("search_for_missing", True)
        random.seed()

        user_selection_allowed = []
        for backend_def in config_dict["backends"]:
            backened_id = backend_def["id"]
            maxpctfull = backend_def.get("max_percent_full", 0)
            weight = backend_def["weight"]
            allow_selection = backend_def.get("allow_selection")
            if allow_selection:
                user_selection_allowed.append(backened_id)
            backend = build_object_store_from_config(config, config_dict=backend_def, fsmon=fsmon)

            self.backends[backened_id] = backend
            self.max_percent_full[backened_id] = maxpctfull

            for _ in range(0, weight):
                # The simplest way to do weighting: add backend ids to a
                # sequence the number of times equalling weight, then randomly
                # choose a backend from that sequence at creation
                self.weighted_backend_ids.append(backened_id)

        self.original_weighted_backend_ids = self.weighted_backend_ids
        self.user_object_store_resolver = user_object_store_resolver
        self.user_selection_allowed = user_selection_allowed
        self.allow_user_selection = bool(user_selection_allowed) or (user_object_store_resolver is not None)
        self.sleeper = None
        if fsmon and (self.global_max_percent_full or [_ for _ in self.max_percent_full.values() if _ != 0.0]):
            self.sleeper = Sleeper()
            self.filesystem_monitor_thread = threading.Thread(target=self.__filesystem_monitor, args=[self.sleeper])
            self.filesystem_monitor_thread.daemon = True
            self.filesystem_monitor_thread.start()
            log.info("Filesystem space monitor started")

    @classmethod
    def parse_xml(clazz, config_xml, legacy=False):
        if legacy:
            backends_root = config_xml
        else:
            backends_root = config_xml.find("backends")

        backends: List[Dict[str, Any]] = []
        config_dict = {
            "search_for_missing": asbool(backends_root.get("search_for_missing", True)),
            "global_max_percent_full": float(backends_root.get("maxpctfull", 0)),
            "backends": backends,
        }

        for b in [e for e in backends_root if e.tag == "backend"]:
            store_id = b.get("id")
            store_weight = int(b.get("weight", 1))
            store_maxpctfull = float(b.get("maxpctfull", 0))
            store_type = b.get("type", "disk")
            store_by = b.get("store_by", None)
            allow_selection = asbool(b.get("allow_selection"))

            objectstore_class, _ = type_to_object_store_class(store_type)
            backend_config_dict = objectstore_class.parse_xml(b)
            backend_config_dict["id"] = store_id
            backend_config_dict["weight"] = store_weight
            backend_config_dict["max_percent_full"] = store_maxpctfull
            backend_config_dict["type"] = store_type
            backend_config_dict["allow_selection"] = allow_selection
            if store_by is not None:
                backend_config_dict["store_by"] = store_by
            backends.append(backend_config_dict)

        return config_dict

    @classmethod
    def from_xml(
        clazz,
        config,
        config_xml,
        fsmon=False,
        user_object_store_resolver: Optional[UserObjectStoreResolver] = None,
        **kwd,
    ):
        legacy = False
        if config_xml is None:
            distributed_config = config.distributed_object_store_config_file
            assert distributed_config is not None, (
                "distributed object store ('object_store = distributed') "
                "requires a config file, please set one in "
                "'distributed_object_store_config_file')"
            )

            log.debug("Loading backends for distributed object store from %s", distributed_config)
            config_xml = parse_xml(distributed_config).getroot()
            legacy = True
        else:
            log.debug("Loading backends for distributed object store from %s", config_xml.get("id"))

        config_dict = clazz.parse_xml(config_xml, legacy=legacy)
        return clazz(config, config_dict, fsmon=fsmon, user_object_store_resolver=user_object_store_resolver)

    def to_dict(self, object_store_uris: Optional[Set[str]] = None) -> Dict[str, Any]:
        as_dict = super().to_dict()
        as_dict["global_max_percent_full"] = self.global_max_percent_full
        as_dict["search_for_missing"] = self.search_for_missing
        backends: List[Dict[str, Any]] = []
        for backend_id, backend in self.backends.items():
            backend_as_dict = backend.to_dict()
            backend_as_dict["id"] = backend_id
            backend_as_dict["max_percent_full"] = self.max_percent_full[backend_id]
            backend_as_dict["weight"] = len([i for i in self.original_weighted_backend_ids if i == backend_id])
            backends.append(backend_as_dict)
        if object_store_uris:
            for user_object_store_uri in object_store_uris:
                if not self.user_object_store_resolver:
                    raise ConfigDoesNotAllowException()

                object_store_config = self.user_object_store_resolver.resolve_object_store_uri_config(
                    user_object_store_uri
                )
                dynamic_object_store_as_dict = user_object_store_configuration_to_config_dict(
                    object_store_config, user_object_store_uri
                )
                backends.append(dynamic_object_store_as_dict)

        as_dict["backends"] = backends
        return as_dict

    def shutdown(self):
        """Shut down. Kill the free space monitor if there is one."""
        super().shutdown()
        if self.sleeper is not None:
            self.sleeper.wake()

    def __filesystem_monitor(self, sleeper: Sleeper):
        while self.running:
            new_weighted_backend_ids = self.original_weighted_backend_ids
            for id, backend in self.backends.items():
                maxpct = self.max_percent_full[id] or self.global_max_percent_full
                pct = backend.get_store_usage_percent()
                if pct > maxpct:
                    new_weighted_backend_ids = [_ for _ in new_weighted_backend_ids if _ != id]
            self.weighted_backend_ids = new_weighted_backend_ids
            sleeper.sleep(120)  # Test free space every 2 minutes

    def _construct_path(self, obj, **kwargs):
        return self._resolve_backend(obj.object_store_id).construct_path(obj, **kwargs)

    def _create(self, obj, **kwargs):
        """The only method in which obj.object_store_id may be None."""
        object_store_id = obj.object_store_id
        if object_store_id is None or not self._exists(obj, **kwargs):
            if object_store_id is None or (object_store_id not in self.backends and "://" not in object_store_id):
                try:
                    object_store_id = random.choice(self.weighted_backend_ids)
                    obj.object_store_id = object_store_id
                except IndexError:
                    raise ObjectInvalid(
                        f"objectstore.create, could not generate obj.object_store_id: {obj}, kwargs: {kwargs}"
                    )
                log.debug(
                    "Selected backend '%s' for creation of %s %s", object_store_id, obj.__class__.__name__, obj.id
                )
            else:
                log.debug(
                    "Using preferred backend '%s' for creation of %s %s",
                    object_store_id,
                    obj.__class__.__name__,
                    obj.id,
                )
            return self._resolve_backend(object_store_id).create(obj, **kwargs)
        else:
            return self._resolve_backend(object_store_id)

    def _call_method(self, method, obj, default, default_is_exception, **kwargs):
        object_store_id = self.__get_store_id_for(obj, **kwargs)
        if object_store_id is not None:
            return self._resolve_backend(object_store_id).__getattribute__(method)(obj, **kwargs)
        if default_is_exception:
            raise default(
                f"objectstore, _call_method failed: {method} on {self._repr_object_for_exception(obj)}, kwargs: {kwargs}"
            )
        else:
            return default

    def _resolve_backend(self, object_store_id: str):
        try:
            return self.backends[object_store_id]
        except KeyError:
            if object_store_id.startswith("user_objects://") and self.user_object_store_resolver:
                return self.user_object_store_resolver.resolve_object_store_uri(object_store_id)
            raise

    def get_quota_source_map(self) -> "QuotaSourceMap":
        if self._quota_source_map is None:
            quota_source_map = QuotaSourceMap()
            self._merge_quota_source_map(quota_source_map, self)
            self._quota_source_map = quota_source_map
        assert self._quota_source_map is not None
        return self._quota_source_map

    def get_device_source_map(self) -> "DeviceSourceMap":
        if self._device_source_map is None:
            device_source_map = DeviceSourceMap()
            self._merge_device_source_map(device_source_map, self)
            self._device_source_map = device_source_map
        assert self._device_source_map is not None
        return self._device_source_map

    @classmethod
    def _merge_quota_source_map(clz, quota_source_map, object_store):
        for backend_id, backend in object_store.backends.items():
            if isinstance(backend, DistributedObjectStore):
                clz._merge_quota_source_map(quota_source_map, backend)
            else:
                quota_source_map.backends[backend_id] = backend.get_quota_source_map()

    @classmethod
    def _merge_device_source_map(clz, device_source_map: "DeviceSourceMap", object_store):
        for backend_id, backend in object_store.backends.items():
            if isinstance(backend, DistributedObjectStore):
                clz._merge_device_source_map(device_source_map, backend)
            else:
                device_source_map.backends[backend_id] = backend.get_device_source_map()

    def __get_store_id_for(self, obj, **kwargs):
        if obj.object_store_id is not None:
            if obj.object_store_id in self.backends or obj.object_store_id.startswith("user_objects://"):
                return obj.object_store_id
            else:
                log.warning(
                    "The backend object store ID (%s) for %s object with ID %s is invalid",
                    obj.object_store_id,
                    obj.__class__.__name__,
                    obj.id,
                )
        elif self.search_for_missing:
            # if this instance has been switched from a non-distributed to a
            # distributed object store, or if the object's store id is invalid,
            # try to locate the object
            for id, store in self.backends.items():
                if store.exists(obj, **kwargs):
                    log.warning(
                        f"{obj.__class__.__name__} object with ID {obj.id} found in backend object store with ID {id}"
                    )
                    obj.object_store_id = id
                    return id
        return None

    def object_store_ids(self, private=None):
        object_store_ids = []
        for backend_id, backend in self.backends.items():
            object_store_ids.extend(backend.object_store_ids(private=private))
            if backend.private is private or private is None:
                object_store_ids.append(backend_id)
        return object_store_ids

    def object_store_allows_id_selection(self) -> bool:
        """Return True if this object store respects object_store_id and allow selection of this."""
        return self.allow_user_selection

    def validate_selected_object_store_id(self, user, object_store_id: Optional[str]) -> Optional[str]:
        parent_check = super().validate_selected_object_store_id(user, object_store_id)
        if parent_check or object_store_id is None:
            return parent_check
        # user selection allowed and object_store_id is not None
        if object_store_id.startswith("user_objects://"):
            if not user:
                return "Supplied object store id is not accessible"
            rest_of_uri = object_store_id.split("://", 1)[1]
            user_object_store_uuid = rest_of_uri
            for user_object_store in user.object_stores:
                if str(user_object_store.uuid) == user_object_store_uuid:
                    return None
            return "Supplied object store id was not found"
        if object_store_id not in self.object_store_ids_allowing_selection():
            return "Supplied object store id is not an allowed object store selection"
        return None

    def object_store_ids_allowing_selection(self) -> List[str]:
        """Return a non-empty list of allowed selectable object store IDs during creation."""
        return self.user_selection_allowed

    def get_concrete_store_by_object_store_id(self, object_store_id: str) -> Optional["ConcreteObjectStore"]:
        """If this is a distributed object store, get ConcreteObjectStore by object_store_id."""
        return self.backends[object_store_id]


class HierarchicalObjectStore(NestedObjectStore):
    """
    ObjectStore that defers to a list of backends.

    When getting objects the first store where the object exists is used.
    When creating objects only the first store is used.
    """

    store_type = "hierarchical"

    def __init__(self, config, config_dict, fsmon=False):
        """The default constructor. Extends `NestedObjectStore`."""
        super().__init__(config, config_dict)

        backends: Dict[int, ObjectStore] = {}
        is_private = config_dict.get("private", DEFAULT_PRIVATE)
        for order, backend_def in enumerate(config_dict["backends"]):
            backend_is_private = backend_def.get("private")
            if backend_is_private is not None:
                assert (
                    is_private == backend_is_private
                ), "The private attribute must be defined on the HierarchicalObjectStore and not contained concrete objectstores."
            backend_quota = backend_def.get("quota")
            if backend_quota is not None:
                # Make sure just was using defaults - because cannot override what is
                # is setup by the HierarchicalObjectStore.
                assert backend_quota.get("source", DEFAULT_QUOTA_SOURCE) == DEFAULT_QUOTA_SOURCE
                assert backend_quota.get("enabled", DEFAULT_QUOTA_ENABLED) == DEFAULT_QUOTA_ENABLED

            backends[order] = build_object_store_from_config(config, config_dict=backend_def, fsmon=fsmon)

        self.backends = backends
        self.private = is_private
        quota_config = config_dict.get("quota", {})
        self.quota_source = quota_config.get("source", DEFAULT_QUOTA_SOURCE)
        self.quota_enabled = quota_config.get("enabled", DEFAULT_QUOTA_ENABLED)

    @classmethod
    def parse_xml(clazz, config_xml):
        backends_list = []
        is_private = BaseObjectStore.parse_private_from_config_xml(config_xml)
        for backend in sorted(config_xml.find("backends"), key=lambda b: int(b.get("order"))):
            store_type = backend.get("type")
            objectstore_class, _ = type_to_object_store_class(store_type)
            backend_config_dict = objectstore_class.parse_xml(backend)
            backend_config_dict["private"] = is_private
            backend_config_dict["type"] = store_type
            backends_list.append(backend_config_dict)

        config_dict = {"backends": backends_list}
        config_dict["private"] = is_private
        return config_dict

    def to_dict(self):
        as_dict = super().to_dict()
        backends = []
        for backend in self.backends.values():
            backend_as_dict = backend.to_dict()
            backends.append(backend_as_dict)
        as_dict["backends"] = backends
        as_dict["private"] = self.private
        return as_dict

    def _exists(self, obj, **kwargs):
        """Check all child object stores."""
        for store in self.backends.values():
            if store.exists(obj, **kwargs):
                return True
        return False

    def _construct_path(self, obj, **kwargs):
        return self.backends[0].construct_path(obj, **kwargs)

    def _create(self, obj, **kwargs):
        """Call the primary object store."""
        return self.backends[0].create(obj, **kwargs)

    def _is_private(self, obj):
        # Unlink the DistributedObjectStore - the HierarchicalObjectStore does not use
        # object_store_id - so all the contained object stores need to define is_private
        # the same way.
        return self.private

    def get_quota_source_map(self):
        quota_source_map = QuotaSourceMap(
            self.quota_source,
            self.quota_enabled,
        )
        return quota_source_map


def serialize_static_object_store_config(object_store: ObjectStore, object_store_uris: Set[str]) -> Dict[str, Any]:
    """Serialize a static object store configuration for database-less serialization.

    The database-less part here comes from the fact these are used in job directories
    during extended metadata collection. Any database/vault/app config details should
    be unpacked and the result should be an object store configuration that doesn't
    depend on those entities but which resolves to the same locations.
    """
    if len(object_store_uris) == 0:
        return object_store.to_dict()
    if not isinstance(object_store, DistributedObjectStore):
        # TODO: Not for the MVP or first iteration - but potentially we could allow
        # a concrete store here and then build a Distributed store from that and
        # concrete stores represented by object_store_uris
        raise ConfigDoesNotAllowException("ObjectStore configuration does not allow per-user object stores")
    return object_store.to_dict(object_store_uris=object_store_uris)


class QuotaModel(BaseModel):
    source: Optional[str] = None
    enabled: bool


class ConcreteObjectStoreModel(BaseModel):
    object_store_id: Optional[str] = None
    private: bool
    name: Optional[str] = None
    description: Optional[str] = None
    quota: QuotaModel
    badges: List[BadgeDict]
    device: Optional[str] = None


def type_to_object_store_class(
    store: str, fsmon: bool = False, user_object_store_resolver: Optional[UserObjectStoreResolver] = None
) -> Tuple[Type[BaseObjectStore], Dict[str, Any]]:
    objectstore_class: Type[BaseObjectStore]
    objectstore_constructor_kwds: Dict[str, Any] = {}
    if store == "disk":
        objectstore_class = DiskObjectStore
    elif store == "boto3":
        from .s3_boto3 import S3ObjectStore as Boto3ObjectStore

        objectstore_class = Boto3ObjectStore
    elif store in ["s3", "aws_s3"]:
        from .s3 import S3ObjectStore

        objectstore_class = S3ObjectStore
    elif store == "cloud":
        from .cloud import Cloud

        objectstore_class = Cloud
    elif store in ["swift", "generic_s3"]:
        from .s3 import GenericS3ObjectStore

        objectstore_class = GenericS3ObjectStore
    elif store == "distributed":
        objectstore_class = DistributedObjectStore
        objectstore_constructor_kwds["fsmon"] = fsmon
        objectstore_constructor_kwds["user_object_store_resolver"] = user_object_store_resolver
    elif store == "hierarchical":
        objectstore_class = HierarchicalObjectStore
        objectstore_constructor_kwds["fsmon"] = fsmon
    elif store == "irods":
        from .irods import IRODSObjectStore

        objectstore_class = IRODSObjectStore
    elif store == "azure_blob":
        from .azure_blob import AzureBlobObjectStore

        objectstore_class = AzureBlobObjectStore
    elif store == "pithos":
        from .pithos import PithosObjectStore

        objectstore_class = PithosObjectStore
    elif store == "rucio":
        from .rucio import RucioObjectStore

        objectstore_class = RucioObjectStore
    elif store == "onedata":
        from .onedata import OnedataObjectStore

        objectstore_class = OnedataObjectStore
    else:
        raise Exception(f"Unrecognized object store definition: {store}")
    # Disable the Pulsar object store for now until it receives some attention
    # elif store == 'pulsar':
    #    from .pulsar import PulsarObjectStore
    #    return PulsarObjectStore(config=config, config_xml=config_xml)

    return objectstore_class, objectstore_constructor_kwds


def build_test_object_store_from_user_config(
    config,
    object_store_config: ObjectStoreConfiguration,
):
    # check an object store configuration by building a standalone object store
    # from a supplied user object store configuration.
    config_dict = user_object_store_configuration_to_config_dict(object_store_config, uuid4().hex)
    object_store = build_object_store_from_config(
        config,
        config_dict=config_dict,
        disable_process_management=True,
    )
    return object_store


def build_object_store_from_config(
    config,
    fsmon=False,
    config_xml=None,
    config_dict=None,
    disable_process_management=False,
    user_object_store_resolver: Optional[UserObjectStoreResolver] = None,
):
    """
    Invoke the appropriate object store.

    Will use the `object_store_config_file` attribute of the `config` object to
    configure a new object store from the specified XML file.

    Or you can specify the object store type in the `object_store` attribute of
    the `config` object. Currently 'disk', 's3', 'swift', 'distributed',
    'hierarchical', 'irods', and 'pulsar' are supported values.
    """
    from_object = "xml"

    if config is None and config_dict is not None and "config" in config_dict:
        # Build an application config object from to_dict of an ObjectStore.
        config = Bunch(disable_process_management=disable_process_management, **config_dict["config"])
    elif config is None:
        raise Exception(
            "build_object_store_from_config sent None as config parameter and one cannot be recovered from config_dict"
        )

    if config_xml is None and config_dict is None:
        config_file = config.object_store_config_file
        if os.path.exists(config_file):
            log.debug("Reading object store config from file: %s", config_file)
            if config_file.endswith(".xml") or config_file.endswith(".xml.sample"):
                # This is a top level invocation of build_object_store_from_config, and
                # we have an object_store_conf.xml -- read the .xml and build
                # accordingly
                config_xml = parse_xml(config.object_store_config_file).getroot()
                store = config_xml.get("type")
            else:
                with open(config_file) as f:
                    config_dict = yaml.safe_load(f)
                from_object = "dict"
                store = config_dict.get("type")
        elif config.object_store_config:
            log.debug("Reading object store config from object_store_config option")
            from_object = "dict"
            config_dict = config.object_store_config
            store = config_dict.get("type")
        else:
            store = config.object_store
    elif config_xml is not None:
        store = config_xml.get("type")
    elif config_dict is not None:
        from_object = "dict"
        store = config_dict.get("type")

    objectstore_class, objectstore_constructor_kwds = type_to_object_store_class(
        store, fsmon=fsmon, user_object_store_resolver=user_object_store_resolver
    )
    if from_object == "xml":
        return objectstore_class.from_xml(config=config, config_xml=config_xml, **objectstore_constructor_kwds)
    else:
        return objectstore_class(config=config, config_dict=config_dict, **objectstore_constructor_kwds)


# View into the application configuration that is shared between the global object store
# and user defined object stores as produced by concrete_object_store.
class UserObjectStoresAppConfig(BaseModel):
    object_store_cache_path: str
    object_store_cache_size: int
    user_config_templates_use_saved_configuration: Literal["fallback", "preferred", "never"]
    jobs_directory: str
    new_file_path: str
    umask: int
    gid: int


# TODO: this will need app details...
# TODO: unit test from configuration dict...
def concrete_object_store(
    object_store_configuration: ObjectStoreConfiguration, app_config: UserObjectStoresAppConfig
) -> ConcreteObjectStore:
    # Adapt structured UserObjectStoresAppConfig into a more full configuration object as expected by
    # the object stores
    class GalaxyConfigAdapter:
        # Hard code these, these will not support legacy features
        object_store_check_old_style = False
        object_store_store_by = "uuid"

        # Set this to false for now... not sure but we may want to revisit this
        enable_quotas = False

        # These need to come in from Galaxy's config
        jobs_directory = app_config.jobs_directory
        new_file_path = app_config.new_file_path
        umask = app_config.umask
        gid = app_config.gid
        object_store_cache_size = app_config.object_store_cache_size
        object_store_cache_path = app_config.object_store_cache_path

    objectstore_class, objectstore_constructor_kwds = type_to_object_store_class(
        store=object_store_configuration.type,
        fsmon=False,
    )
    assert issubclass(objectstore_class, ConcreteObjectStore)
    return objectstore_class(
        config=GalaxyConfigAdapter(),
        config_dict=object_store_configuration.model_dump(),
        **objectstore_constructor_kwds,
    )


def local_extra_dirs(func):
    """Non-local plugin decorator using local directories for the extra_dirs (job_work and temp)."""

    def wraps(self, *args, **kwargs):
        if kwargs.get("base_dir", None) is None:
            return func(self, *args, **kwargs)
        else:
            for c in self.__class__.__mro__:
                if c.__name__ == "DiskObjectStore":
                    return getattr(c, func.__name__)(self, *args, **kwargs)
            raise Exception(
                f"Could not call DiskObjectStore's {func.__name__} method, does your "
                "Object Store plugin inherit from DiskObjectStore?"
            )

    return wraps


def config_to_dict(config):
    """Dict-ify the portion of a config object consumed by the ObjectStore class and its subclasses."""
    return {
        "object_store_check_old_style": config.object_store_check_old_style,
        "enable_quotas": config.enable_quotas,
        "file_path": config.file_path,
        "umask": config.umask,
        "jobs_directory": config.jobs_directory,
        "new_file_path": config.new_file_path,
        "object_store_cache_path": config.object_store_cache_path,
        "object_store_cache_size": config.object_store_cache_size,
        "gid": config.gid,
    }


class QuotaSourceInfo(NamedTuple):
    label: Optional[str]
    use: bool


class DeviceSourceMap:
    def __init__(self, device_id=DEFAULT_DEVICE_ID):
        self.default_device_id = device_id
        self.backends = {}

    def get_device_id(self, object_store_id: str) -> Optional[str]:
        if object_store_id in self.backends:
            device_map = self.backends.get(object_store_id)
            if device_map:
                print(device_map)
                return device_map.get_device_id(object_store_id)

        return self.default_device_id


class QuotaSourceMap:
    def __init__(self, source=DEFAULT_QUOTA_SOURCE, enabled=DEFAULT_QUOTA_ENABLED):
        self.default_quota_source = source
        self.default_quota_enabled = enabled
        self.info = QuotaSourceInfo(self.default_quota_source, self.default_quota_enabled)
        self.backends = {}
        self._labels = None

    def get_quota_source_info(self, object_store_id):
        if object_store_id in self.backends:
            return self.backends[object_store_id].get_quota_source_info(object_store_id)
        else:
            return self.info

    def get_quota_source_label(self, object_store_id):
        if object_store_id in self.backends:
            return self.backends[object_store_id].get_quota_source_label(object_store_id)
        else:
            return self.default_quota_source

    def get_quota_source_labels(self):
        if self._labels is None:
            labels = set()
            if self.default_quota_source:
                labels.add(self.default_quota_source)
            for backend in self.backends.values():
                labels = labels.union(backend.get_quota_source_labels())
            self._labels = labels
        return self._labels

    def default_usage_excluded_ids(self):
        exclude_object_store_ids = []
        for backend_id, backend_source_map in self.backends.items():
            if backend_source_map.default_quota_source is not None:
                exclude_object_store_ids.append(backend_id)
            elif not backend_source_map.default_quota_enabled:
                exclude_object_store_ids.append(backend_id)
        return exclude_object_store_ids

    def get_id_to_source_pairs(self, include_default_quota_source=False):
        pairs = []
        for backend_id, backend_source_map in self.backends.items():
            if (
                backend_source_map.default_quota_source is not None or include_default_quota_source
            ) and backend_source_map.default_quota_enabled:
                pairs.append((backend_id, backend_source_map.default_quota_source))
        return pairs

    def ids_per_quota_source(self, include_default_quota_source=False):
        quota_sources: Dict[Optional[str], List[str]] = {}
        for object_id, quota_source_label in self.get_id_to_source_pairs(
            include_default_quota_source=include_default_quota_source
        ):
            if quota_source_label not in quota_sources:
                quota_sources[quota_source_label] = []
            quota_sources[quota_source_label].append(object_id)
        return quota_sources


class ObjectCreationProblem(Exception):
    pass


# Calling these client_message to make it clear they should use the language of the
# Galaxy UI/UX - for instance "storage location" not "objectstore".
class ObjectCreationProblemSharingDisabled(ObjectCreationProblem):
    client_message = "Job attempted to create sharable output datasets in a storage location with sharing disabled"


class ObjectCreationProblemStoreFull(ObjectCreationProblem):
    client_message = (
        "Job attempted to create output datasets in a full storage location, please contact your admin for more details"
    )


class ObjectStorePopulator:
    """Small helper for interacting with the object store and making sure all
    datasets from a job end up with the same object_store_id.
    """

    def __init__(self, has_object_store, user):
        if hasattr(has_object_store, "object_store"):
            object_store = has_object_store.object_store
        else:
            object_store = has_object_store
        self.object_store = object_store
        self.object_store_id = None
        self.user = user

    def set_object_store_id(self, data, require_shareable=False):
        self.set_dataset_object_store_id(data.dataset, require_shareable=require_shareable)

    def set_dataset_object_store_id(self, dataset, require_shareable=True):
        # Create an empty file immediately.  The first dataset will be
        # created in the "default" store, all others will be created in
        # the same store as the first.
        dataset.object_store_id = self.object_store_id
        try:
            ensure_non_private = require_shareable
            concrete_store = self.object_store.create(dataset, ensure_non_private=ensure_non_private)
            if concrete_store.private and require_shareable:
                raise ObjectCreationProblemSharingDisabled()
        except ObjectInvalid:
            raise ObjectCreationProblemStoreFull()
        self.object_store_id = dataset.object_store_id  # these will be the same thing after the first output


def persist_extra_files(
    object_store: ObjectStore,
    src_extra_files_path: str,
    primary_data: "DatasetInstance",
    extra_files_path_name: Optional[str] = None,
) -> None:
    if not primary_data.dataset.purged and os.path.exists(src_extra_files_path):
        assert primary_data.dataset
        if not extra_files_path_name:
            extra_files_path_name = primary_data.dataset.extra_files_path_name_from(object_store)
        assert extra_files_path_name
        persist_extra_files_for_dataset(object_store, src_extra_files_path, primary_data.dataset, extra_files_path_name)


def persist_extra_files_for_dataset(
    object_store: ObjectStore,
    src_extra_files_path: str,
    dataset: "Dataset",
    extra_files_path_name: str,
):
    for root, _dirs, files in safe_walk(src_extra_files_path):
        extra_dir = os.path.join(extra_files_path_name, os.path.relpath(root, src_extra_files_path))
        extra_dir = os.path.normpath(extra_dir)
        for f in files:
            if not in_directory(f, src_extra_files_path):
                # Unclear if this can ever happen if we use safe_walk ... probably not ?
                raise MalformedContents(f"Invalid dataset path: {f}")
            object_store.update_from_file(
                dataset,
                extra_dir=extra_dir,
                alt_name=f,
                file_name=os.path.join(root, f),
                create=True,
                preserve_symlinks=True,
            )



"""
Object Store plugin for the Microsoft Azure Block Blob Storage system
"""

import logging
import os
from datetime import (
    datetime,
    timedelta,
)

try:
    from azure.common import AzureHttpError
    from azure.storage.blob import (
        BlobSasPermissions,
        BlobServiceClient,
        generate_blob_sas,
    )
except ImportError:
    BlobServiceClient = None  # type: ignore[assignment,unused-ignore,misc]

from ._caching_base import CachingConcreteObjectStore
from .caching import (
    enable_cache_monitor,
    parse_caching_config_dict_from_xml,
)

NO_BLOBSERVICE_ERROR_MESSAGE = (
    "ObjectStore configured, but no azure.storage.blob dependency available."
    "Please install and properly configure azure.storage.blob or modify Object Store configuration."
)

log = logging.getLogger(__name__)


def parse_config_xml(config_xml):
    try:
        auth_xml = config_xml.findall("auth")[0]

        account_name = auth_xml.get("account_name")
        account_key = auth_xml.get("account_key")
        account_url = auth_xml.get("account_url")

        container_xml = config_xml.find("container")
        container_name = container_xml.get("name")

        transfer_xml = config_xml.findall("transfer")
        if not transfer_xml:
            transfer_xml = {}
        else:
            transfer_xml = transfer_xml[0]
        transfer_dict = {}
        for key in [
            "max_concurrency",
            "download_max_concurrency",
            "upload_max_concurrency",
            "max_single_put_size",
            "max_single_get_size",
            "max_block_size",
        ]:
            value = transfer_xml.get(key)
            if transfer_xml.get(key) is not None:
                transfer_dict[key] = value

        cache_dict = parse_caching_config_dict_from_xml(config_xml)

        tag, attrs = "extra_dir", ("type", "path")
        extra_dirs = config_xml.findall(tag)
        if not extra_dirs:
            msg = f"No {tag} element in XML tree"
            log.error(msg)
            raise Exception(msg)
        extra_dirs = [{k: e.get(k) for k in attrs} for e in extra_dirs]
        auth = {
            "account_name": account_name,
            "account_key": account_key,
        }
        if account_url:
            auth["account_url"] = account_url

        config_dict = {
            "auth": auth,
            "container": {
                "name": container_name,
            },
            "cache": cache_dict,
            "transfer": transfer_dict,
            "extra_dirs": extra_dirs,
            "private": CachingConcreteObjectStore.parse_private_from_config_xml(config_xml),
        }
        name = config_xml.attrib.get("name", None)
        if name is not None:
            config_dict["name"] = name
        device = config_xml.attrib.get("device", None)
        config_dict["device"] = device
        return config_dict
    except Exception:
        # Toss it back up after logging, we can't continue loading at this point.
        log.exception("Malformed ObjectStore Configuration XML -- unable to continue")
        raise


class AzureBlobObjectStore(CachingConcreteObjectStore):
    """
    Object store that stores objects as blobs in an Azure Blob Container. A local
    cache exists that is used as an intermediate location for files between
    Galaxy and Azure.
    """

    store_type = "azure_blob"
    cloud = True

    def __init__(self, config, config_dict):
        super().__init__(config, config_dict)

        auth_dict = config_dict["auth"]
        container_dict = config_dict["container"]
        cache_dict = config_dict.get("cache") or {}
        self.enable_cache_monitor, self.cache_monitor_interval = enable_cache_monitor(config, config_dict)

        self.account_name = auth_dict.get("account_name")
        self.account_url = auth_dict.get("account_url")
        self.account_key = auth_dict.get("account_key")

        self.container_name = container_dict.get("name")
        raw_transfer_dict = config_dict.get("transfer", {})
        typed_transfer_dict = {}
        for key in [
            "max_concurrency",
            "download_max_concurrency",
            "upload_max_concurrency",
            "max_single_put_size",
            "max_single_get_size",
            "max_block_size",
        ]:
            value = raw_transfer_dict.get(key)
            if value is not None:
                typed_transfer_dict[key] = int(value)
        self.transfer_dict = typed_transfer_dict

        self.cache_size = cache_dict.get("size") or self.config.object_store_cache_size
        self.staging_path = cache_dict.get("path") or self.config.object_store_cache_path
        self.cache_updated_data = cache_dict.get("cache_updated_data", True)

        self._initialize()

    def _initialize(self):
        if BlobServiceClient is None:
            raise Exception(NO_BLOBSERVICE_ERROR_MESSAGE)

        self._configure_connection()
        self._ensure_staging_path_writable()
        self._start_cache_monitor_if_needed()

    def to_dict(self):
        as_dict = super().to_dict()
        auth = {
            "account_name": self.account_name,
            "account_key": self.account_key,
        }
        if self.account_url:
            auth["account_url"] = self.account_url
        as_dict.update(
            {
                "auth": auth,
                "container": {
                    "name": self.container_name,
                },
                "transfer": self.transfer_dict,
                "cache": {
                    "size": self.cache_size,
                    "path": self.staging_path,
                    "cache_updated_data": self.cache_updated_data,
                },
            }
        )
        return as_dict

    # config_xml is an ElementTree object.
    @classmethod
    def parse_xml(clazz, config_xml):
        return parse_config_xml(config_xml)

    def _configure_connection(self):
        log.debug("Configuring Connection")
        extra_kwds = {}
        for key in [
            "max_single_put_size",
            "max_single_get_size",
            "max_block_size",
        ]:
            if key in self.transfer_dict:
                extra_kwds[key] = self.transfer_dict[key]

        if self.account_url:
            # https://pypi.org/project/azure-storage-blob/
            service = BlobServiceClient(
                account_url=self.account_url,
                credential={"account_name": self.account_name, "account_key": self.account_key},
                **extra_kwds,
            )
        else:
            service = BlobServiceClient(
                account_url=f"https://{self.account_name}.blob.core.windows.net",
                credential=self.account_key,
                **extra_kwds,
            )
        self.service = service

    def _get_remote_size(self, rel_path):
        try:
            properties = self._blob_client(rel_path).get_blob_properties()
            size_in_bytes = properties.size
            return size_in_bytes
        except AzureHttpError:
            log.exception("Could not get size of blob '%s' from Azure", rel_path)
            return -1

    def _blobs_from(self, rel_path):
        return self.service.get_container_client(self.container_name).list_blobs(name_starts_with=rel_path)

    def _exists_remotely(self, rel_path: str):
        try:
            is_dir = rel_path[-1] == "/"
            if is_dir:
                blobs = self._blobs_from(rel_path)
                if blobs:
                    return True
                else:
                    return False
            else:
                exists = self._blob_client(rel_path).exists()
        except AzureHttpError:
            log.exception("Trouble checking existence of Azure blob '%s'", rel_path)
            return False
        return exists

    def _blob_client(self, rel_path: str):
        return self.service.get_blob_client(self.container_name, rel_path)

    def _download(self, rel_path):
        local_destination = self._get_cache_path(rel_path)
        try:
            log.debug("Pulling '%s' into cache to %s", rel_path, local_destination)
            if not self._caching_allowed(rel_path):
                return False
            else:
                self._download_to_file(rel_path, local_destination)
                return True
        except AzureHttpError:
            log.exception("Problem downloading '%s' from Azure", rel_path)
        return False

    def _download_to_file(self, rel_path, local_destination):
        kwd = {}
        max_concurrency = self.transfer_dict.get("download_max_concurrency") or self.transfer_dict.get(
            "max_concurrency"
        )
        if max_concurrency is not None:
            kwd["max_concurrency"] = max_concurrency
        with open(local_destination, "wb") as f:
            self._blob_client(rel_path).download_blob().download_to_stream(f, **kwd)

    def _download_directory_into_cache(self, rel_path, cache_path):
        blobs = self._blobs_from(rel_path)
        for blob in blobs:
            key = blob.name
            local_file_path = os.path.join(cache_path, os.path.relpath(key, rel_path))

            # Create directories if they don't exist
            os.makedirs(os.path.dirname(local_file_path), exist_ok=True)

            # Download the file
            self._download_to_file(key, local_file_path)

    def _push_string_to_path(self, rel_path: str, from_string: str) -> bool:
        try:
            self._blob_client(rel_path).upload_blob(from_string, overwrite=True)
            return True
        except AzureHttpError:
            log.exception("Trouble pushing to Azure Blob '%s' from string", rel_path)
            return False

    def _push_file_to_path(self, rel_path: str, source_file: str) -> bool:
        try:
            with open(source_file, "rb") as f:
                kwd = {}
                max_concurrency = self.transfer_dict.get("upload_max_concurrency") or self.transfer_dict.get(
                    "max_concurrency"
                )
                if max_concurrency is not None:
                    kwd["max_concurrency"] = max_concurrency
                self._blob_client(rel_path).upload_blob(f, overwrite=True, **kwd)
            return True
        except AzureHttpError:
            log.exception("Trouble pushing to Azure Blob '%s' from file '%s'", rel_path, source_file)
            return False

    def _delete_remote_all(self, rel_path: str) -> bool:
        try:
            blobs = self._blobs_from(rel_path)
            for blob in blobs:
                log.debug("Deleting from Azure: %s", blob)
                self._blob_client(blob.name).delete_blob()
            return True
        except AzureHttpError:
            log.exception("Could not delete blob '%s' from Azure", rel_path)
            return False

    def _delete_existing_remote(self, rel_path: str) -> bool:
        try:
            self._blob_client(rel_path).delete_blob()
            return True
        except AzureHttpError:
            log.exception("Could not delete blob '%s' from Azure", rel_path)
            return False

    def _get_object_url(self, obj, **kwargs):
        if self._exists(obj, **kwargs):
            rel_path = self._construct_path(obj, **kwargs)
            try:
                url = self._blob_client(rel_path).url
                # https://learn.microsoft.com/en-us/azure/storage/blobs/sas-service-create-python
                token = generate_blob_sas(
                    account_name=self.account_name,
                    account_key=self.account_key,
                    container_name=self.container_name,
                    blob_name=rel_path,
                    permission=BlobSasPermissions(read=True),
                    expiry=datetime.utcnow() + timedelta(hours=1),
                )
                return f"{url}?{token}"
            except AzureHttpError:
                log.exception("Trouble generating URL for dataset '%s'", rel_path)
        return None

    def _get_store_usage_percent(self, obj):
        # Percent used for Azure blob containers is effectively zero realistically.
        # https://learn.microsoft.com/en-us/azure/storage/blobs/scalability-targets
        return 0.0

    def shutdown(self):
        self._shutdown_cache_monitor()
